/*
 * Standard C-library includes.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

/*
 * Unix includes.
 */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/*
 * Local includes.
 */
#include "ccbconstants.h"
#include "ccbserverlink.h"

typedef struct {
  CCBServerLink *link;      /* The communications interface of the server */
  int shutdown;             /* This flag is set to true to tell the server's */
                            /*  event loop to terminate. */
  int reboot;               /* This flag tells the server to restart itself */
  struct timeval timeout;   /* The timeout argument of select() */
  unsigned long ninteg;     /* The number of integrations that have been */
                            /*  started so far within the current scan. */
  unsigned long nmonitor;   /* The number of monitor messages sent so far */
  unsigned long scan;       /* The ID of the current scan */
  CCBConfig *cnf;           /* The CCB configuration of the current scan */
  CCBTimeStamp wakeup_time; /* The time of the next simulated event */
  int driver_loaded;        /* True when the device-driver is loaded */
  int scan_active;          /* True while a scan or intra-scan is running */
} CCBDemoServer;

static CCBDemoServer *new_CCBDemoServer(void);
static CCBDemoServer *del_CCBDemoServer(CCBDemoServer *cds);
static CCB_DRIVER_LOAD_FN(demo_driver_load_fn);
static CCB_DRIVER_UNLOAD_FN(demo_driver_unload_fn);
static CCB_DRIVER_TELL_FN(demo_driver_tell_fn);
static CCB_REBOOT_RTC_FN(demo_reboot_rtc_fn);
static CCB_SHUTDOWN_RTC_FN(demo_shutdown_rtc_fn);

static int demo_server_loop(CCBDemoServer *cds);
static int demo_queue_fake_integration(CCBDemoServer *cds);
static int demo_queue_fake_monitoring(CCBDemoServer *cds);
static int demo_display_config(CCBDemoServer *cds);
static int demo_prepare_scan(CCBDemoServer *cds, unsigned long scan,
			     unsigned long mjd, unsigned long tod);
static int demo_config_scan(CCBDemoServer *cds, const CCBDrvConfScan *cmd);
static int demo_abort_scan(CCBDemoServer *cds);

/*
 * Specify how long simulated reboots should take.
 */
#define DEMO_REBOOT_DT 10  /* seconds */

/*.......................................................................
 * This is a simple program which illustrates the basic functionality
 * of the server end of the CCB communications library. All it does is
 * allow a manager to connect, report the contents of any messages
 * that it receives from the manager, and periodically send dummy
 * telemetry messages to the manager.
 *
 * Arguments: none
 */
int main(int argc, char *argv[])
{
  CCBDemoServer *cds;  /* The resources of the CCB server */
  int waserr = 0;      /* The completion status */
  int reboot = 0;      /* If true, fork a new server before terminating */
/*
 * In order to simulate the effect of a reboot, we run the server
 * program in a loop.
 */
  do {
/*
 * When restarting after a reboot command, insert a delay to better
 * simulate a reboot.
 */
    if(reboot) {
      printf("demo: Simulated reboot in progress - please wait (%ds).\n",
	     DEMO_REBOOT_DT);
      sleep(DEMO_REBOOT_DT);
    };
/*
 * Report when the server actually starts.
 */
    printf("demo: The CCB server is now starting.\n");
/*
 * Establish the server.
 */
    cds = new_CCBDemoServer();
    if(!cds)
      return 1;
/*
 * Run the server loop.
 */
    waserr = demo_server_loop(cds);
/*
 * Keep a record of whether a reboot was requested.
 */
    reboot = cds->reboot;
/*
 * Shutdown the server and release its resources.
 */
    cds = del_CCBDemoServer(cds);
  } while(reboot);
/*
 * Report termination.
 */
  printf("demo: The CCB server is now terminating.\n");
  return waserr;
}

/*.......................................................................
 * Create a new CCBDemoServer object. This contains all of the resources
 * needed by the demonstration CCB server.
 *
 * Output:
 *  return  CCBDemoServer *  The new object, or NULL on error.
 */
static CCBDemoServer *new_CCBDemoServer(void)
{
  CCBDemoServer *cds;  /* The object to be returned */
/*
 * Allocate the container.
 */
  cds = malloc(sizeof(CCBDemoServer));
  if(!cds) {
    fprintf(stderr, "new_CCBDemoServer: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBDemoServer().
 */
  cds->link = NULL;
  cds->shutdown = 0;
  cds->reboot = 0;
  cds->ninteg = 0;
  cds->nmonitor = 0;
  cds->scan = 0;
  cds->cnf = NULL;
  cds->wakeup_time.mjd = 0;
  cds->wakeup_time.sec = 0;
  cds->wakeup_time.ns = 0;
  cds->driver_loaded = 0;
  cds->scan_active = 0;
/*
 * Create the communications interface of the server.
 */
  cds->link = new_CCBServerLink(cds, demo_driver_load_fn, demo_driver_unload_fn,
				demo_driver_tell_fn, demo_reboot_rtc_fn,
				demo_shutdown_rtc_fn);
  if(!cds->link)
    return del_CCBDemoServer(cds);
/*
 * Allocate a CCB configuration object.
 */
  cds->cnf = new_CCBConfig();
  if(!cds->cnf)
    return del_CCBDemoServer(cds);
  return cds;
}

/*.......................................................................
 * Delete a CCBDemoServer object.
 *
 * Input:
 *  cds     CCBDemoServer *  The object to be deleted.
 * Output:
 *  return  CCBDemoServer *  The deleted object (always NULL).
 */
static CCBDemoServer *del_CCBDemoServer(CCBDemoServer *cds)
{
  if(cds) {
/*
 * Delete the communications interface.
 */
    cds->link = del_CCBServerLink(cds->link);
/*
 * Delete the CCB configuration object.
 */
    cds->cnf = del_CCBConfig(cds->cnf);
/*
 * Delete the container of the object.
 */
    free(cds);
  };
  return NULL;
}

/*.......................................................................
 * Load the simulated CCB device driver.
 */
static CCB_DRIVER_LOAD_FN(demo_driver_load_fn)
{
  CCBDemoServer *cds = (CCBDemoServer *) data;
/*
 * This function isn't supposed to be called until any previously installed
 * driver has been unloaded.
 */
  if(cds->driver_loaded) {
    fprintf(stderr, "demo: Unexpected request to load the CCB driver.\n");
    ccb_record_server_faults(cds->link, CCB_SOFT_FAULT);
    return 1;
  };
/*
 * Mark the device driver as now loaded.
 */
  cds->driver_loaded = 1;
  printf("demo: CCB device-driver loaded.\n");
  return 0;
}

/*.......................................................................
 * Unload the simulated CCB device driver.
 */
static CCB_DRIVER_UNLOAD_FN(demo_driver_unload_fn)
{
  CCBDemoServer *cds = (CCBDemoServer *) data;
/*
 * Mark the device driver as now unloaded.
 */
  cds->driver_loaded = 0;
  printf("demo: CCB device-driver unloaded.\n");
  return 0;
}

/*.......................................................................
 * Execute a device-driver command.
 */
static CCB_DRIVER_TELL_FN(demo_driver_tell_fn)
{
  CCBDemoServer *cds = (CCBDemoServer *) data;
/*
 * This function isn't supposed to be called when the device driver
 * isn't loaded.
 */
  if(!cds->driver_loaded) {
    fprintf(stderr,
	    "demo: Unexpected command while device driver not loaded.\n");
    ccb_record_server_faults(cds->link, CCB_SOFT_FAULT);
    return 1;
  };
/*
 * Determine what type of command has been sent, and execute it.
 */
  switch(cmd->type) {
  case CCB_DRV_INTR_MASK:  /* Set the hardware interrupt mask */
    printf("demo: Integration interrupts %sabled. 1PPS interrupts %sabled.\n",
	   cmd->pars.intr.enable & CCB_INTEG_INTR ? "en":"dis",
	   cmd->pars.intr.enable & CCB_1PPS_INTR ? "en":"dis");
    break;
  case CCB_DRV_LINE_MASK:  /* Mask control line drivers */
    printf(
    "demo: Cal-diode outputs: [A is %smasked] [B is %smasked].\n",
	   cmd->pars.line.lines & CCB_DIODE_A_LINE ? "not ":"",
	   cmd->pars.line.lines & CCB_DIODE_B_LINE ? "not ":"");
    printf(
    "demo: Phase-switch outputs: [A is %smasked] [B is %smasked].\n",
	   cmd->pars.line.lines & CCB_DIODE_A_LINE ? "not ":"",
	   cmd->pars.line.lines & CCB_DIODE_B_LINE ? "not ":"");
    break;
  case CCB_DRV_ABORT_SCAN: /* Halt any ongoing scan or intra-scan */
    if(demo_abort_scan(cds))
      return 1;
    break;
  case CCB_DRV_CONF_SCAN:  /* Configure the next scan */
    if(demo_config_scan(cds, &cmd->pars.conf))
      return 1;
    break;
  case CCB_DRV_STAGE_SCAN: /* Start a new scan at a given time */
    if(demo_prepare_scan(cds, cmd->pars.stage.scan, cmd->pars.stage.mjd,
			 cmd->pars.stage.secs))
      return 1;
    break;
  case CCB_DRV_INTRA_SCAN: /* Start a new intra-scan ASAP */
    if(demo_prepare_scan(cds, cmd->pars.intra.scan, 0, 0))
      return 1;
    break;
  default:
    fprintf(stderr, "demo: Unknown driver command received.\n");
    return 1;
  };
  return 0;
}

/*.......................................................................
 * Stage a simulated reboot.
 */
static CCB_REBOOT_RTC_FN(demo_reboot_rtc_fn)
{
  CCBDemoServer *cds = (CCBDemoServer *) data;
  cds->reboot = 1;
  printf("demo: Reboot requested.\n");
  return 0;
}

/*.......................................................................
 * Stage a simulated shutdown.
 */
static CCB_SHUTDOWN_RTC_FN(demo_shutdown_rtc_fn)
{
  CCBDemoServer *cds = (CCBDemoServer *) data;
  cds->shutdown = 1;
  printf("demo: Shutdown requested.\n");
  return 0;
}

/*.......................................................................
 * This is the I/O-driven event loop that is used to watch for and react
 * to client communications.
 *
 * Input:
 *  cds    CCBDemoServer *  The resources of the server.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int demo_server_loop(CCBDemoServer *cds)
{
/*
 * The event loop runs until the program is either killed, an error
 * causes a return from this function, a shutdown command is received,
 * or a reboot command is received.
 */
  while(!cds->shutdown && !cds->reboot) {
    fd_set rfds;    /* The set of file-descriptors to watch for readability */
    fd_set wfds;    /* The set of file-descriptors to watch for writability */
    int maxfd;      /* The maximum of the file-descriptors to watch */
    int nready;     /* The number of file descriptors ready for I/O */
    CCBInterval dt; /* The timeout duration */
    struct timeval timeout; /* The select() timeout */
/*
 * Construct the sets of file descriptors to be watched from scratch.
 */
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
/*
 * Add the file-descriptors that the server wants us to watch, to rfds
 * and wfds.
 */
    if(ccb_server_select_args(cds->link, &rfds, &wfds, &maxfd))
      return 1;
/*
 * Compute the amount of time remaining until the next wakeup time.
 */
    if(ccb_time_until(cds->wakeup_time, &dt))
      return 1;
/*
 * Record the new select() timeout.
 */
    timeout.tv_sec = dt.sec;
    timeout.tv_usec = dt.ns / CCB_NS_PER_USEC;
/*
 * Wait I/O events and event-triggering timeouts.
 */
    nready = select(maxfd + 1, &rfds, &wfds, NULL,
		    cds->scan_active ? &timeout : NULL);
/*
 * Error?
 */
    if(nready < 0) {
      if(errno != EINTR) {
	fprintf(stderr, "select() error (%s).\n", strerror(errno));
	return 1;
      };
/*
 * Timeout?
 */
    } else if(nready == 0) {
/*
 * Did we just reach the end of an integration period (as opposed to
 * reaching the end of the time spent waiting for the start time of
 * a new scan)?
 */
      if(cds->ninteg > 0) {
/*
 * Queue a fake integration-data message to be sent to the manager.
 */
	if(demo_queue_fake_integration(cds))
	  return 1;
/*
 * Queue a fake monitoring packet.
 */
	if(demo_queue_fake_monitoring(cds))
	  return 1;
      };
/*
 * Start computing the time at which select() should next timeout,
 * by specifying the current time as the wakeup time.
 */
      if(ccb_get_timestamp(&cds->wakeup_time))
	return 1;
/*
 * Compute the duration of the next integration.
 */
      if(ccb_integration_duration(cds->cnf, cds->ninteg, &dt))
	return 1;
/*
 * Set the time at which select() should next timeout, to the current
 * time plus the duration of the next integration.
 */
      if(ccb_add_to_timestamp(&cds->wakeup_time, &dt, &cds->wakeup_time))
	return 1;
/*
 * Increment the count of the number of integrations that have been
 * started within the current scan.
 */
      cds->ninteg++;
/*
 * Report on the duration of the first integration.
 */
      if(cds->ninteg == 1) {
	printf("demo: Duration of first integration = %g seconds.\n",
	       (double) dt.sec + (double) dt.ns / CCB_NS_PER_SEC);
      };
/*
 * At least one file-descriptor is ready for I/O.
 */
    } else {
/*
 * Perform the types of I/O indicated by select().
 */
      if(ccb_server_communicate(cds->link, &rfds, &wfds))
	return 1;
    };
  };
/*
 * Shutdown requested.
 */
  return 0;
}

/*.......................................................................
 * Print out the contents of a CCB configuration object.
 *
 * Input:
 *  cds    CCBDemoServer *  The resources of the server.
 * Output:
 *  return            int     0 - OK.
 *                            1 - Error.
 */
static int demo_display_config(CCBDemoServer *cds)
{
  CCBPhaseSwitchCnf phase;   /* The CCB phase-switch configuration */
  CCBCalDiodeCnf cal;        /* The CCB calibration diode configuration */
  CCBTimingCnf timing;       /* The CCB timing configuration */
  int i;
/*
 * Look up the configuration parameters.
 */
  if(ccb_get_phase_switch_cnf(cds->cnf, &phase) ||
     ccb_get_cal_diode_cnf(cds->cnf, &cal) ||
     ccb_get_timing_cnf(cds->cnf, &timing))
    return 1;
/*
 * Display the phase-switch configuration.
 */
  printf("       Phase-switch configuration:\n"
         "        active_switches=%hu, driven_switches=%hu,\n"
         "        closed_switches=%hu, samp_per_state=%u\n",
	 phase.active_switches, phase.driven_switches,
	 phase.closed_switches, phase.samp_per_state);
/*
 * Display the cal-diode configuration.
 */
  printf("       Cal-diode configuration:\n"
	 "        ncal=%u, driven_diodes=%u,\n"
         "        diode_states={", cal.ncal, cal.driven_diodes);
  for(i=0; i<cal.ncal; i++)
    printf("%hu%s", cal.diode_states[i], i<cal.ncal-1 ? ", " : "");
  printf("}\n"
         "        diode_times={");
  for(i=0; i<cal.ncal; i++)
    printf("%lu%s", cal.diode_times[i], i<cal.ncal-1 ? ", " : "");
  printf("}\n");
/*
 * Display the timing configuration.
 */
  printf("       Timing configuration:\n"
         "        sample_dt=%u, phase_switch_dt=%u,\n"
         "        analog_reset_dt=%u, diode_rise_dt=%lu,\n"
         "        diode_fall_dt=%lu, integ_period=%lu)\n",
	 timing.sample_dt, timing.phase_switch_dt, timing.analog_reset_dt,
	 timing.diode_rise_dt, timing.diode_fall_dt, timing.integ_period);
  return 0;
}

/*.......................................................................
 * Queue a fake integration message to be sent to the manager.
 *
 * Input:
 *  cds    CCBDemoServer *  The resources of the server.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int demo_queue_fake_integration(CCBDemoServer *cds)
{
  unsigned long values[CCB_MAX_INTEG]; /* The fake integrations */
  CCBTimeStamp t;                      /* The current time */
  int i;
/*
 * Lookup the current time.
 */
  if(ccb_get_timestamp(&t))
    return 1;
/*
 * Fill the array of integrated values.
 */
  for(i=0; i<CCB_MAX_INTEG; i++)
    values[i] = UINT_MAX * (double)rand()/RAND_MAX;
/*
 * Queue the integration message.
 */
  if(ccb_queue_integ_msg(cds->link, &t, cds->scan, cds->ninteg-1, values,
			 CCB_MAX_INTEG))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a fake monitoring message to be sent to the manager.
 *
 * Input:
 *  cds    CCBDemoServer *  The resources of the server.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int demo_queue_fake_monitoring(CCBDemoServer *cds)
{
  unsigned long values[CCB_MAX_MONITOR]; /* The fake monitoring values */
  CCBTimeStamp t;                        /* The current time */
  int i;
/*
 * Lookup the current time.
 */
  if(ccb_get_timestamp(&t))
    return 1;
/*
 * Fill the array of monitoring values.
 */
  for(i=0; i<CCB_MAX_MONITOR; i++)
    values[i] = UINT_MAX * (double)rand()/RAND_MAX;
/*
 * Queue the monitor-data message.
 */
  if(ccb_queue_monitor_msg(cds->link, &t, cds->scan, cds->nmonitor++,
			   values, CCB_MAX_MONITOR))
    return 1;
  return 0;
}

/*.......................................................................
 * Prepare the server to start a new scan or intra-scan.
 *
 * Input:
 *  cds    CCBDemoServer *  The resources of the server.
 *  scan   unsigned long    The numeric identifyier of the new scan.
 *  mjd    unsigned long    If none zero, this specifies the date
 *                          at which to start the scan, expressed as
 *                          a Modified Julian Day number.
 *  tod    unsigned long    If mjd!=0, this is the time of day at which
 *                          to start the scan, specified as the number
 *                          of seconds into the day.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int demo_prepare_scan(CCBDemoServer *cds, unsigned long scan,
			     unsigned long mjd, unsigned long tod)
{
  CCBInterval dt;   /* The duration of the delay preceding the integration */
/*
 * Get the name by which to refer to the requested type of scan.
 */
  const char *name = mjd==0 ? "intra-scan" : "scan";
/*
 * Start-scan and intra-scan commands aren't supposed to be invoked
 * when the driver is unloaded.
 */
  if(!cds->driver_loaded) {
    fprintf(stderr, "demo: New %s requested while driver not loaded.\n", name);
    return 1;
  };
/*
 * Start-scan and intra-scan commands aren't supposed to be invoked
 * until any previous scan or intra-scan has been aborted.
 */
  if(cds->scan_active) {
    fprintf(stderr,"demo: Not starting %s, because a scan is already active.\n",
	    name);
    return 1;
  };
/*
 * Report what we are doing.
 */
  printf("demo: Preparing to start %s %lu.\n", name, scan);
/*
 * Record the scan ID counter and reset the integration and monitor counters.
 */
  cds->scan = scan;
  cds->ninteg = 0;
  cds->nmonitor = 0;
  cds->scan_active = 1;
/*
 * Do we need to wait before starting the scan?
 */
  if(mjd != 0) {
    cds->wakeup_time.mjd = mjd;
    cds->wakeup_time.sec = tod;
    cds->wakeup_time.ns = 0;
  } else {  /* Specify the current time as the next select() wakeup time */
    if(ccb_get_timestamp(&cds->wakeup_time))
      return 1;
  };
/*
 * Compute how long we will need to wait before starting the new scan.
 */
  if(ccb_time_until(cds->wakeup_time, &dt))
    return 1;
/*
 * Report how long we have to wait before starting the new scan.
 */
  printf("demo: Waiting %g seconds until start-time of %s.\n",
	 (double) dt.sec + (double) dt.ns / CCB_NS_PER_SEC, name);
  return 0;
}

/*.......................................................................
 * Reconfigure the simulated device driver.
 *
 * Input:
 *  cds         CCBDemoServer *  The resources of the server.
 *  cmd  const CCBDrvConfScan *  The new configuration.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error.
 */ 
static int demo_config_scan(CCBDemoServer *cds, const CCBDrvConfScan *cmd)
{
/*
 * Configuration commands aren't supposed to be invoked when the
 * driver is unloaded.
 */
  if(!cds->driver_loaded) {
    fprintf(stderr,
	    "demo: Scan reconfiguration requested while driver not loaded.\n");
    return 1;
  };
/*
 * This function is only suposed to be called after the most recent scan
 * has been aborted.
 */
  if(cds->scan_active) {
    fprintf(stderr, "demo: Rejecting re-configuration during scan.\n");
    return 1;
  };
/*
 * Copy the configuration into a CCB configuration object.
 */
  if(ccb_set_config(cds->cnf, &cmd->phase, &cmd->cal, &cmd->timing)) {
    fprintf(stderr, "demo: Configuration error.\n");
    return 1;
  };
/*
 * Report the new configuration.
 */
  printf("demo: Installing new scan configuration:\n");
  if(demo_display_config(cds))
    return 1;
  return 0;
}

/*.......................................................................
 * Abort the current scan or intra-scan, if any.
 *
 * Input:
 *  cds         CCBDemoServer *  The resources of the server.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error.
 */ 
static int demo_abort_scan(CCBDemoServer *cds)
{
/*
 * Abort-scan commands aren't supposed to be invoked when the
 * driver isn't loaded.
 */
  if(!cds->driver_loaded) {
    fprintf(stderr, "demo: Scan abort requested while driver not loaded.\n");
    return 1;
  };
/*
 * Mark the current scan/intra-scan as stopped.
 */
  cds->scan_active = 0;
  return 0;
}
