/* Standard C includes */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* Local includes */

#include "ccbclientlink.h"
#include "ccbclientpriv.h"
#include "ccbconstants.h"
#include "ccbtelemetryclient.h"
#include "ccbcontrolclient.h"
#include "ccbprint.h"

/*
 * Collect all callback functions and callback data.
 */
typedef struct {
  CCBClientSocketsFn *client_sockets_fn; /* The socket event-handler */
                                         /*  callback */
  void *client_sockets_data;             /* Callback data of */
                                         /*  client_sockets_fn(). */
} CCBClientLinkCallbacks;

/*
 * The following constant is designed to be assigned to any
 * CCBClientLinkCallbacks structure to set all of its contained
 * function pointers and data pointers to their respective versions of
 * nul. This takes advantage of the fact that the C standard states
 * that statically allocated structures are initialized as though all
 * of their members were individually assigned 0 (NB this isn't the
 * same as setting all bytes to zero when pointers are involved).  Not
 * only does this tremendously reduce the number of statements needed
 * to initialize one of these structures in the CCBClientLink
 * constructor, but it is also safer, because it makes it impossible
 * to forget to initialize a newly added callback and its data
 * pointer.
 */
static const CCBClientLinkCallbacks init_client_callbacks;

/*
 * A container of the following type is used to keep a record of the
 * details sent to the application by the last invokation of the
 * client_sockets callback. This is used to avoid calling this callback
 * redundantly.
 */
typedef struct {
  int cntrl_sock;       /* The control socket */
  int telem_sock;       /* The telemetry socket */
  unsigned io;          /* The set of I/O events to watch for */
} CCBClientSocketsCache;

/*
 * CCBClientLink objects manage the resources associated with a connection
 * to a remote (or local) CCB server. Note that a typedef alias for
 * this datatype is defined in ccbclientlink.h.
 */
struct CCBClientLink {
  CCBTelemetryClient *tc;           /* The telemetry client */
  CCBControlClient *cc;             /* The control client */
  CCBClientLinkCallbacks cb;        /* Callback functions */
  unsigned long ping_count;         /* The number of ping commands sent */
                                    /*  so far. */
  unsigned ping_echos;              /* A bitwise union of CCBLinkType */
                                    /*  enumerators, recording replies */
                                    /*  to the last ping command. */
  CCBClientSocketsCache sock_cache; /* The cache of the last info sent by */
                                    /*  the client_socket callback function. */
};

static int ccb_client_error(CCBClientLink *cl);

/*
 * Return the types of I/O that the client should be watching for,
 * expressed as a bitwise union of CCBClientIOStatus enumerators.
 */
static unsigned ccb_client_io_status(CCBClientLink *cl);

/*
 * Report socket file descriptors and the events that they are awaiting.
 */
static int ccb_report_client_sockets(CCBClientLink *cl, int force);

/*.......................................................................
 * Return a CCBClientLink communications object for the manager to use to
 * communicate with the CCB server on a specified host.
 *
 * Input:
 *  errdim         size_t    The allocated size of the errmsg[] buffer.
 * Input/Output:
 *  errmsg           char *  When, upon error, new_CCBClientLink() returns
 *                           NULL, an error message is composed in the
 *                           buffer pointed to by the errmsg argument.
 *                           Where needed, this is truncated to fit within
 *                           the first errdim elements, in which case,
 *                           unless errdim is 0, errmsg[errdim-1] is
 *                           assigned '\0', to ensure that the string
 *                           is correctly terminated.
 *                           Alternatively, If new_CCBClientLink() returns
 *                           a CCBClientLink object, the contents of errmsg[]
 *                           are left untouched, and subsequent
 *                           error messages from other functions are
 *                           reported as log messages via the log message
 *                           callback (see ccb_log_msg_callback()).
 * Output:
 *  return  CCBClientLink *  The resource object needed to communicate with
 *                           the specified CCB server, or NULL on error.
 */
CCBClientLink *new_CCBClientLink(size_t errdim, char *errmsg)
{
  CCBClientLink *cl;   /* The new object */
/*
 * Check the arguments.
 */
  if(!errmsg) {
    fprintf(stderr, "new_CCBClientLink: NULL error message buffer.\n");
    errno = EINVAL;
    return NULL;
  };
/*
 * Allocate the container of the new object.
 */
  cl = (CCBClientLink *) malloc(sizeof(CCBClientLink));
  if(!cl) {
    ccb_snprintf(errmsg, errdim, "Insufficient memory for CCBClientLink object");
    errno = ENOMEM;
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to a point at which it can safely be passed
 * to del_CCBClientLink().
 */
  cl->tc = NULL;
  cl->cc = NULL;
  cl->cb = init_client_callbacks;
  cl->ping_count = 0;
  cl->ping_echos = 0;
  cl->sock_cache.cntrl_sock = -1;
  cl->sock_cache.telem_sock = -1;
  cl->sock_cache.io = 0;
/*
 * Allocate the object which manages the telemetry link.
 */
  cl->tc = new_CCBTelemetryClient(cl);
  if(!cl->tc)
    return del_CCBClientLink(cl);
/*
 * Allocate the object which manages the telemetry link.
 */
  cl->cc = new_CCBControlClient(cl);
  if(!cl->cc)
    return del_CCBClientLink(cl);
/*
 * The client resources are now ready for use.
 */
  return cl;
}

/*.......................................................................
 * Disconnect from the CCB server associated with a given CCBClientLink
 * object, and return the resources of the object to the system.
 *
 * Input:
 *  cl       CCBClientLink *  The client side of the telemetry and control
 *                            links to the CCB.
 * Output:
 *  return   CCBClientLink *  Always NULL (ie. the deleted object).
 */
CCBClientLink *del_CCBClientLink(CCBClientLink *cl)
{
  if(cl) {
/*
 * Delete the resources of the telemetry and control clients.
 */
    cl->tc = del_CCBTelemetryClient(cl->tc);
    cl->cc = del_CCBControlClient(cl->cc);
/*
 * Delete the now empty container.
 */
    free(cl);
  };
  return NULL;
}

/*.......................................................................
 * Initiate a connection to a new CCB server.
 *
 * Input:
 *  cl  CCBClientLink *  The client side of the telemetry and control
 *                       links to the CCB.
 *  host   const char *  The address of the CCB server to contact.
 *                       When the 'nonblocking' argument is zero,
 *                       this can be either a numeric IP address or
 *                       a host name. When the 'nonblocking' argument
 *                       is non-zero, only numeric IP addresses are
 *                       accepted, since name-server lookups take
 *                       time.
 *  nonblocking   int    When this argument is 0, ccb_client_connect().
 *                       doesn't return until the control and
 *                       telemetry links have either been established or
 *                       rejected. When non-zero, non-blocking connects are
 *                       initiated to the control and telemetry
 *                       servers, and this function may return before
 *                       these connections have completed. When this
 *                       happens, responding to the completion or
 *                       rejection of the connections is left to
 *                       subsequent calls to ccb_client_communicate().
 * Output:
 *  return        int    0 - OK.
 *                       1 - Error.
 */
int ccb_client_connect(CCBClientLink *cl, const char *host, int nonblocking)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(103),
	"ccb_client_connect: NULL argument(s).\n");    
    return 1;
  };
/*
 * Terminate any existing connection.
 */
  ccb_client_disconnect(cl);
/*
 * Forward the connection request to the control and telemetry link
 * controllers.
 */
  if(ccb_cc_connect(cl->cc, host, nonblocking) ||
     ccb_tc_connect(cl->tc, host, nonblocking))
    return ccb_client_error(cl);
/*
 * Zero the ping counter, and mark both links as verified, until the
 * next ping is sent.
 */
  cl->ping_count = 0;
  cl->ping_echos = CCB_ALL_LINKS;
/*
 * Report the new sockets to the application.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Disconnect a CCB manager-client from the CCB server that it is currently
 * talking to.
 *
 * Input:
 *  cl       CCBClientLink *  The client side of the telemetry and control
 *                            links to the CCB.
 */
void ccb_client_disconnect(CCBClientLink *cl)
{
  if(cl) {
    ccb_cc_disconnect(cl->cc);  /* Disconnect the control client */
    ccb_tc_disconnect(cl->tc);  /* Disconnect the telemetry client */
/*
 * Report the socket closures to the application.
 */
    ccb_report_client_sockets(cl, 0);
  };
}

/*.......................................................................
 * Toggle whether non-blocking socket I/O is used for communications
 * with a specified CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  on                  int    0 - Switch to blocking I/O.
 *                             1 - Switch to non-blocking I/O.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_client_non_blocking_io(CCBClientLink *cl, int on)
{
  int status = 0;  /* The return status of this function */
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(105),
	"ccb_client_non_blocking_io: NULL argument(s).\n");    
    return 1;
  };
/*
 * Forward the request to the individual client objects.
 */
  if(ccb_cc_non_blocking_io(cl->cc, on))
    status = 1;
  if(ccb_tc_non_blocking_io(cl->tc, on))
    status = 1;
  return status;
}

/*.......................................................................
 * Return the types of I/O that the client should be watching for,
 * expressed as a bitwise union of CCBClientIOStatus enumerators.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 * Output:
 *  return         unsigned    A bitwise union of CCBClientIOStatus
 *                             enumerators (see ccbclientlink.h), indicating
 *                             what forms of I/O the caller should be
 *                             watching for.
 */
static unsigned ccb_client_io_status(CCBClientLink *cl)
{
  unsigned status = 0;                   /* The return value */
  CCBControlClientStatus control_status; /* The status of the control link */
  CCBTelemClientStatus telem_status;     /* The status of the telemetry link */
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(106),
	"ccb_client_io_status: NULL argument(s).\n");    
    return status;
  };
/*
 * Get the statuses of the control and telemetry links.
 */
  control_status = ccb_cc_status(cl->cc);
  telem_status = ccb_tc_status(cl->tc);
/*
 * Do we have a control socket?
 */
  if(control_status != CCB_CONTROL_CLIENT_DOWN) {
/*
 * The client should always be watching the control socket for
 * readability, so as to see connection completion events, incoming
 * command replies, and connection loss.
 */
    status |= CCB_CNTRL_READ;
/*
 * If there are any incompletely sent, or queued outgoing control
 * commands, or we are awaiting the completion of a non-blocking
 * connect, tell the caller to watch the writability of the control
 * socket.
 */
    if(control_status == CCB_CONTROL_CLIENT_BUSY ||
       control_status == CCB_CONTROL_CLIENT_PEND)
      status |= CCB_CNTRL_WRITE;
  };
/*
 * Do we have a telemetry socket?
 */
  if(telem_status != CCB_TELEM_CLIENT_DOWN) {
/*
 * The client should always be watching the telemetry socket for
 * readability, so as to see connection completion events, incoming
 * command replies, and connection loss.
 */
    status |= CCB_TELEM_READ;
/*
 * While waiting for the completion of a non-blocking connect, the
 * client must watch for writability of the telemetry socket.
 */
    if(telem_status == CCB_TELEM_CLIENT_PEND)
      status |= CCB_TELEM_WRITE;
  };
/*
 * Return the accumulated set of I/O events to watch for.
 */
  return status;
}

/*.......................................................................
 * Perform the types of socket I/O listed in the bitwise union of
 * CCBClientIOStatus values in the io argument.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  io             unsigned    The bitwise union of I/O operations to
 *                             perform.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_client_communicate(CCBClientLink *cl, unsigned io)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(107),
	"ccb_client_communicate: NULL argument(s).\n");    
    return 1;
  };
/*
 * Read from the control socket?
 */
  if(io & CCB_CNTRL_READ) {
    if(ccb_cc_read_control(cl->cc))
      return ccb_client_error(cl);
/*
 * Write to the control socket?
 */
  } else if(io & CCB_CNTRL_WRITE) {
    if(ccb_cc_send_control(cl->cc))
      return ccb_client_error(cl);
/*
 * Read from the telemetry socket?
 */
  } else if(io & CCB_TELEM_READ) {
    if(ccb_tc_read_telemetry(cl->tc))
      return ccb_client_error(cl);
/*
 * Telemetry socket writable?
 */
  } else if(io & CCB_TELEM_WRITE) {
    if(ccb_tc_send_telemetry(cl->tc))
      return ccb_client_error(cl);
  };
/*
 * Update the application's idea of what events need to be watched for.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Register a command-error callback function with a given CCB client
 * connection.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  fn        CCBCmdErrorFn *  The function to call whenever a control
 *                             command evokes an error report.
 *  data               void *  Anonymous data to pass to fn().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cmd_error_callback(CCBClientLink *cl, CCBCmdErrorFn *fn, void *data)
{
/*
 * Check the arguments.
 */
  if(!cl || !fn) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(108),
	"ccb_cmd_error_callback: NULL argument(s).\n");
    return 1;
  };
  return ccb_cc_cmd_error_callback(cl->cc, fn, data);
}

/*.......................................................................
 * Queue a start-scan command to be sent to a given CCB server.
 *
 * Input:
 *  cl     CCBClientLink *  The client side of the telemetry and control
 *                          links to the CCB.
 *  id              long    An identification number to use to
 *                          refer to this command if the server
 *                          rejects it for some reason.
 *  cnf  const CCBConfig *  The CCB configuration to associate with the
 *                          new scan.
 *  scan   unsigned long    The numeric ID to give the new scan.
 *  mjd         unsigned    This is the date at which the scan should be
 *                          started, expressed in UTC, as a Modified
 *                          Julian Day number.  To be precise, this is
 *                          the integer part of (Julian_Date - 2400000.5). 
 *  tod         unsigned    This is the time of day at which the scan
 *                          should be started, specified as the integer
 *                          number of seconds after 0H UTC on the day
 *                          indicated by 'mjd'. The scan starts at the
 *                          start of the specified second, provided that
 *                          the command is received at least one second
 *                          in advance of this time.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
int ccb_queue_start_scan_cmd(CCBClientLink *cl, long id, const CCBConfig *cnf,
                             unsigned long scan, unsigned long mjd,
			     unsigned long tod)
{
/*
 * Check the arguments.
 */
  if(!cl || !cnf) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(113),
	"ccb_queue_start_scan_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  return ccb_cc_queue_start_scan_cmd(cl->cc, id, cnf, scan, mjd, tod);
}

/*.......................................................................
 * Queue a stop-scan command to be sent to a given CCB server.
 *
 * Input:
 *  cl     CCBClientLink *  The client side of the telemetry and control
 *                          links to the CCB.
 *  id              long    An identification number to use to
 *                          refer to this command if the server
 *                          rejects it for some reason.
 *  cnf  const CCBConfig *  The CCB configuration to associate with the
 *                          new scan.
 *  scan   unsigned long    The numeric ID to give the new intra-scan.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
int ccb_queue_stop_scan_cmd(CCBClientLink *cl, long id, const CCBConfig *cnf,
			    unsigned long scan)
{
/*
 * Check the arguments.
 */
  if(!cl || !cnf) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(114),
	"ccb_queue_stop_scan_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_stop_scan_cmd(cl->cc, id, cnf, scan))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a reset command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_reset_cmd(CCBClientLink *cl, long id)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(115),
	"ccb_queue_reset_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_reset_cmd(cl->cc, id))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a standby command to be sent to a given CCB server.
 *
 * Input:
 *  cl            CCBClientLink *  The client side of the telemetry and
 *                                 control links to the CCB.
 *  id                     long    An identification number to use to
 *                                 refer to this command if the server
 *                                 rejects it for some reason.
 *  stream_mask  unsigned short    This parameter contains a bit-wise
 *                                 union of CCBTelemetryStream enumerators, 
 *                                 specifying which of the currently enabled
 *                                 telemetry streams should continue to be
 *                                 sent to the manager. To determine this,
 *                                 the CCB server takes the union of the
 *                                 set of streams specified in this
 *                                 parameter with the set previously
 *                                 specified by the last CCBTelemetryCnf
 *                                 command. 
 * Output:
 *  return                  int    0 - OK.
 *                                 1 - Error.
 */
int ccb_queue_standby_cmd(CCBClientLink *cl, long id,
			  unsigned short stream_mask)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(116),
	"ccb_queue_standby_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_standby_cmd(cl->cc, id, stream_mask))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue an awaken command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_awaken_cmd(CCBClientLink *cl, long id)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(117),
	"ccb_queue_awaken_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_awaken_cmd(cl->cc, id))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a ping command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_ping_cmd(CCBClientLink *cl, long id)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(118),
	"ccb_queue_ping_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Record the fact that the latest ping hasn't had any replies yet.
 */
  cl->ping_echos = 0;
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_ping_cmd(cl->cc, id, ++cl->ping_count))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a check-status command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_status_request_cmd(CCBClientLink *cl, long id)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(119),
	"ccb_queue_status_request_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_status_request_cmd(cl->cc, id))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a shutdown command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_shutdown_cmd(CCBClientLink *cl, long id)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(120),
	"ccb_queue_shutdown_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_shutdown_cmd(cl->cc, id))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a reboot command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_reboot_cmd(CCBClientLink *cl, long id)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(121),
	"ccb_queue_reboot_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_reboot_cmd(cl->cc, id))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a monitoring-control command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 *  period   unsigned short    The monitoring period, expressed as a
 *                             multiple of the integration period.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_monitor_cmd(CCBClientLink *cl, long id, unsigned short period)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(224),
	"ccb_queue_monitor_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_monitor_cmd(cl->cc, id, period))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a telemetry-control command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 *  streams  unsigned short    The set of telemetry streams to report to
 *                             the manager, expressed as a bitwise union
 *                             of CCBTelemetryStream enumerators.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_telemetry_cmd(CCBClientLink *cl, long id, unsigned short streams)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(225),
	"ccb_queue_telemetry_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_telemetry_cmd(cl->cc, id, streams))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a log-control command to be sent to a given CCB server.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 *  period    unsigned long    The time interval at which the
 *                             log-server forgets about historically
 *                             sent messages (seconds).
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_queue_logger_cmd(CCBClientLink *cl, long id, unsigned long period)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(247),
	"ccb_queue_logger_cmd: NULL argument(s).\n");    
    return cl ? ccb_cc_command_error(cl->cc, id, CCB_CMD_SYSERR) : 1;
  };
/*
 * Forward the request to the control client.
 */
  if(ccb_cc_queue_logger_cmd(cl->cc, id, period))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Register the callback function to be called whenever a given CCB client
 * object receives a control-status reply message.
 *
 * Input:
 *  cl         CCBClientLink *  The client side of the telemetry and control
 *                              links to the CCB.
 *  fn  CCBStatusReplyFn *  The function to call whenever a reply
 *                              to a previously sent control-status
 *                              command is received over the control link.
 *  data                void *  Anonymous data to pass to fn().
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_status_reply_callback(CCBClientLink *cl, CCBStatusReplyFn *fn,
				   void *data)
{
  if(!cl || !fn) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(123),
	"ccb_status_reply_callback: NULL argument(s).\n");    
    return 1;
  };
  return ccb_cc_status_reply_callback(cl->cc, fn, data);
}

/*.......................................................................
 * Register the callback function to be called whenever a given CCB client
 * object receives a monitor data message.
 *
 * Input:
 *  cl    CCBClientLink *  The client side of the telemetry and control
 *                         links to the CCB.
 *  fn  CCBMonitorMsgFn *  The function to call whenever a monitor-
 *                         data message is received over the
 *                         telemetry link.
 *  data           void *  Anonymous data to pass to fn().
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
int ccb_monitor_msg_callback(CCBClientLink *cl, CCBMonitorMsgFn *fn, 
			     void *data)
{
  if(!cl || !fn) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(124),
	"ccb_monitor_msg_callback: NULL argument(s).\n");    
    return 1;
  };
  return ccb_tc_monitor_msg_callback(cl->tc, fn, data);
}

/*.......................................................................
 * Register the callback function to be called whenever a given CCB client
 * object receives an integrated-data message.
 *
 * Input:
 *  cl    CCBClientLink *  The client side of the telemetry and control
 *                         links to the CCB.
 *  fn    CCBIntegMsgFn *  The function to call whenever an
 *                         integration-data message is received
 *                         over the telemetry link.
 *  data           void *  Anonymous data to pass to fn().
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
int ccb_integ_msg_callback(CCBClientLink *cl, CCBIntegMsgFn *fn,
				void *data)
{
  if(!cl || !fn) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(125),
	"ccb_integ_msg_callback: NULL argument(s).\n");    
    return 1;
  };
  return ccb_tc_integ_msg_callback(cl->tc, fn, data);
}

/*.......................................................................
 * Register the callback function to be called whenever a given CCB client
 * object receives a log message.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  fn          CCBLogMsgFn *  The function to call whenever a
 *                             log message is received, either from the
 *                             server over the telemetry link, or from the
 *                             local end of the communications library.
 *  data               void *  Anonymous data to pass to fn().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_log_msg_callback(CCBClientLink *cl, CCBLogMsgFn *fn, void *data)
{
  if(!cl || !fn) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(126),
	"ccb_log_msg_callback: NULL argument(s).\n");
    return 1;
  };
  return ccb_tc_log_msg_callback(cl->tc, fn, data);
}

/*.......................................................................
 * Close both the control and telemetry connections to the CCB server,
 * in response to an error, then return non-zero to allow this
 * function to be used in return statements that need to return a
 * non-zero error code.
 *
 * Input:
 *  cl    CCBClientLink *  The networking interface of the CCB client.
 * Output:
 *  return          int    Always 1 (ie. the error return code of most
 *                         functions in this module).
 */
static int ccb_client_error(CCBClientLink *cl)
{
/*
 * Disconnect the telemetry and control clients.
 */
  ccb_client_disconnect(cl);
/*
 * Queue a log message to be sent to the next manager that connects to
 * the telemetry client.
 */
  ccb_log_client_msg(cl, CCB_WARNING, CCB_LOGID(131),
    "The connection to the CCB server is now closed.\n");
/*
 * Return the normal error code of most functions in this module.
 */
  return 1;
}

/*.......................................................................
 * This is a printf-style function that queues a formatted log message for
 * subsequent dispatch to the manager.
 *
 * Input:
 *  cl  CCBClientLink *  The communications resource object of the CCB
 *                       client, or NULL to write to stderr.
 *  level CCBLogLevel    The significance of the message.
 *  id  unsigned long    The numeric identifier of the message. This must
 *                       be chosen to be unique within the CCB software,
 *                       and be the lowest so-far unused id number >= 0.
 *  fmt    const char *  This is a standard printf-style format string.
 *  ...                  The arguments refered to by the format specifiers
 *                       in the fmt[] argument.
 * Output:
 *  return        int    0 - OK.
 *                       1 - Error.
 */
int ccb_log_client_msg(CCBClientLink *cl, CCBLogLevel level, unsigned long id,
		       const char *fmt, ...)
{
  int status;          /* The return value of ccb_tc_log_client_msg() */
  va_list ap;          /* The encapsulated variable argument list */
/*
 * Encapsulate the variable argument list and pass the result to the
 * telemetry client, where the log-reporting callback is stored.
 */
  va_start(ap, fmt);
  status = ccb_tc_vlog_client_msg(cl ? cl->tc : NULL,level,id,fmt,ap);
  va_end(ap);
  return status;
}

/*.......................................................................
 * This is a vfprintf-style function that queues a formatted log message for
 * subsequent dispatch to the manager.
 *
 * Input:
 *  cl  CCBClientLink *  The communications resource object of the CCB
 *                       client, or NULL to write to stderr.
 *  level CCBLogLevel    The significance of the message.
 *  id  unsigned long    The numeric identifier of the message. This must
 *                       be chosen to be unique within the CCB software,
 *                       and be the lowest so-far unused id number >= 0.
 *  fmt    const char *  This is a standard printf-style format string.
 *  ap        va_list    The encapsulated list of arguments refered to
 *                       by the format specifiers in the fmt[] argument.
 * Output:
 *  return        int    0 - OK.
 *                       1 - Error.
 */
int ccb_vlog_client_msg(CCBClientLink *cl, CCBLogLevel level, unsigned long id,
			const char *fmt, va_list ap)
{
/*
 * Queue the log message for dispatch.
 */
  if(ccb_tc_vlog_client_msg(cl ? cl->tc : NULL, level,id,fmt,ap))
    return 1;
/*
 * Queuing this command may have changed the I/O events that should be
 * awaited. Inform the application of any changes.
 */
  if(ccb_report_client_sockets(cl, 0))
    return 1;
  return 0;
}

/*.......................................................................
 * Add to given select() file-descriptor sets according to the I/O
 * currently being waited for by *cl.
 * 
 *
 * Input:
 *  cl    CCBClientLink *  The networking interface of a CCB client.
 * Input/Output:
 *  rfds         fd_set *  The socket file-descriptors that should be
 *                         watched for readability will be added to the
 *                         existing contents of this set.
 *  wfds         fd_set *  The socket file-descriptors that should be
 *                         watched for writability will be added to the
 *                         existing contents of this set.
 *  maxfd           int *  On input this should be the maximum of all
 *                         file descriptors that the caller has placed
 *                         in rfds and wfds. On output this will have
 *                         been updated to be the maximum of this number
 *                         and the maximum file descriptor added by this
 *                         function.
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
int ccb_client_select_args(CCBClientLink *cl, fd_set *rfds, fd_set *wfds,
			   int *maxfd)
{
  int cntrl_sock;  /* The current control-socket */
  int telem_sock; /* The current telemetry socket */
  unsigned io;    /* The set of I/O events being awaited */
/*
 * Check the arguments.
 */
  if(!cl || !rfds || !wfds || !maxfd) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(202),
	"ccb_client_select_args: NULL argument(s).\n");    
    return 1;
  };
/*
 * Lookup the set of I/O events that are currently being awaited by *cl.
 */
  io = ccb_client_io_status(cl);
/*
 * Lookup the current socket file-descriptors.
 */
  cntrl_sock = ccb_cc_socket(cl->cc);
  telem_sock = ccb_tc_socket(cl->tc);
/*
 * Transpose the contents of the I/O set to the corresponding select()
 * file-descriptor sets.
 */
  if(cntrl_sock >= 0) {
    if(io & CCB_CNTRL_READ)
      FD_SET(cntrl_sock, rfds);
    if(io & CCB_CNTRL_WRITE)
      FD_SET(cntrl_sock, wfds);
    if(cntrl_sock > *maxfd)
      *maxfd = cntrl_sock;
  };
  if(telem_sock >= 0) {
    if(io & CCB_TELEM_READ)
      FD_SET(telem_sock, rfds);
    if(io & CCB_TELEM_WRITE)
      FD_SET(telem_sock, wfds);
    if(telem_sock > *maxfd)
      *maxfd = telem_sock;
  };
  return 0;
}

/*.......................................................................
 * Fill in two poll() file-descriptor event structures according to
 * the I/O that is currently being awaited by *cl.
 * 
 * Input:
 *  cl    CCBClientLink *  The networking interface of a CCB client.
 *  size            int    The allocated size of the fds[] array.
 * Input/Output:
 *  fds   struct pollfd *  The array of poll() file-descriptor event
 *                         structures. This must have at least 2 free
 *                         elements. On output up to 2 file descriptors
 *                         are added to this array, placed after the
 *                         initial *nfds currently used entries.
 *  nfds            int *  On input *nfd must specify how many elements
 *                         at the start of fds[] are currently occupied.
 *                         On output this number is updated to include
 *                         the number of file descriptors (up to 2)
 *                         added to fds[] by this function.
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
int ccb_client_poll_args(CCBClientLink *cl, int size, struct pollfd *fds,
			 int *nfds)
{
  int cntrl_sock;  /* The current control-socket */
  int telem_sock; /* The current telemetry socket */
  unsigned io;    /* The set of I/O events being awaited */
/*
 * Check the arguments.
 */
  if(!cl || !fds || !nfds) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(250),
	"ccb_client_select_args: NULL argument(s).\n");    
    return 1;
  };
/*
 * Make sure that *fds has at least 2 unused elements.
 */
  if(size < 0 || *nfds < 0 || *nfds + 2 > size) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(251),
	"ccb_client_poll_args: fds[] array too small.\n");    
    return 1;
  };
/*
 * Lookup the set of I/O events that are currently being awaited by *cl.
 */
  io = ccb_client_io_status(cl);
/*
 * Lookup the current socket file-descriptors.
 */
  cntrl_sock = ccb_cc_socket(cl->cc);
  telem_sock = ccb_tc_socket(cl->tc);
/*
 * Transpose the contents of the I/O set to the corresponding select()
 * file-descriptor sets.
 */
  if(cntrl_sock >= 0) {
    struct pollfd *fd = fds + (*nfds)++;
    fd->fd = cntrl_sock;
    fd->events = 0;
    if(io & CCB_CNTRL_READ)
      fd->events |= POLLIN;
    if(io & CCB_CNTRL_WRITE)
      fd->events |= POLLOUT;
  };
  if(telem_sock >= 0) {
    struct pollfd *fd = fds + (*nfds)++;
    fd->fd = telem_sock;
    fd->events = 0;
    if(io & CCB_TELEM_READ)
      fd->events |= POLLIN;
    if(io & CCB_TELEM_WRITE)
      fd->events |= POLLOUT;
  };
  return 0;
}

/*.......................................................................
 * Convert the contents of select() file-descriptor sets to an I/O mask
 * of the form expected by ccb_client_communicate().
 *
 * Input:
 *  cl    CCBClientLink *  The networking interface of a CCB client.
 *  rfds   const fd_set *  The set of file-descriptors that should be
 *                         watched for readability.
 *  wfds   const fd_set *  The set of file-descriptors that should be
 *                         watched for writability.
 * Output:
 *  return     unsigned    The bitmask union of CCBClientIOStatus
 *                         enumerators that correspond to the contents
 *                         of rfds and wfds.
 */
unsigned ccb_client_selected_io(CCBClientLink *cl, const fd_set *rfds,
				const fd_set *wfds)
{
  unsigned io;    /* The bitmask set of CCBClientIOStatus enumerators */
  int cntrl_sock;  /* The current control-socket */
  int telem_sock; /* The current telemetry socket */
/*
 * Check the arguments.
 */
  if(!cl || !rfds || !wfds) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(203),
	"ccb_client_selected_io: NULL argument(s).\n");    
    return 0;  /* No events */
  };
/*
 * Lookup the current socket file-descriptors.
 */
  cntrl_sock = ccb_cc_socket(cl->cc);
  telem_sock = ccb_tc_socket(cl->tc);
/*
 * Translate the contents of rfds and wfds to the corresponding bitmask
 * union of CCBClientIOStatus enumerators.
 */
  io = 0;
  if(cntrl_sock >= 0) {
    if(FD_ISSET(cntrl_sock, rfds))
      io |= CCB_CNTRL_READ;
    if(FD_ISSET(cntrl_sock, wfds))
      io |= CCB_CNTRL_WRITE;
  };
  if(telem_sock >= 0) {
    if(FD_ISSET(telem_sock, rfds))
      io |= CCB_TELEM_READ;
    if(FD_ISSET(telem_sock, wfds))
      io |= CCB_TELEM_WRITE;
  };
/*
 * Return the set of I/O events.
 */
  return io;
}

/*.......................................................................
 * Convert the contents of poll() file-descriptor sets to an I/O mask of
 * the form expected by ccb_client_communicate().
 *
 * Input:
 *  cl         CCBClientLink *  The networking interface of a CCB client.
 *  fds  const struct pollfd *  The array of poll file-descriptor event
 *                              structures that contains the sub-array of
 *                              entries that ccb_client_poll_args()
 *                              filled in and that poll() subsequently
 *                              updated. Entries containing unknown file
 *                              descriptors are silently ignored.
 *  nfds                 int    The number of elements in fds[].
 * Output:
 *  return          unsigned    The bitmask union of CCBClientIOStatus
 *                              enumerators that correspond to the contents
 *                              of fds[0..nfds-1].
 */
unsigned ccb_client_polled_io(CCBClientLink *cl, const struct pollfd *fds,
			      int nfds)
{
  unsigned io;    /* The bitmask set of CCBClientIOStatus enumerators */
  int cntrl_sock;  /* The current control-socket */
  int telem_sock; /* The current telemetry socket */
  int i;
/*
 * Check the arguments.
 */
  if(!cl || !fds) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(252),
	"ccb_client_polled_io: NULL argument(s).\n");    
    return 0;  /* No events */
  };
/*
 * Lookup the current socket file-descriptors.
 */
  cntrl_sock = ccb_cc_socket(cl->cc);
  telem_sock = ccb_tc_socket(cl->tc);
/*
 * Translate the contents of rfds and wfds to the corresponding bitmask
 * union of CCBClientIOStatus enumerators.
 */
  io = 0;
/*
 * Go through the contents of the array, looking for our file descriptors
 * and the corresponding I/O events that poll indicated.
 */
  for(i=0; i<nfds; i++) {
    const struct pollfd *pfd = fds + i;
    short revents = pfd->events;
/*
 * Control socket events.
 */
    if(pfd->fd == cntrl_sock && cntrl_sock >= 0) {
      if(revents & POLLIN)
	io |= CCB_CNTRL_READ;
      if(revents & POLLOUT)
	io |= CCB_CNTRL_WRITE;
/*
 * Telemetry socket events.
 */
    } else if(pfd->fd == telem_sock && telem_sock >= 0) {
      if(revents & POLLIN)
	io |= CCB_TELEM_READ;
      if(revents & POLLOUT)
	io |= CCB_TELEM_WRITE;
    };
  };
/*
 * Return the set of I/O events.
 */
  return io;
}

/*.......................................................................
 * Report the receipt of a ping reply from a given link.
 *
 * Input:
 *  cl         CCBClientLink *  The networking interface of a CCB client.
 *  type         CCBLinkType    The link over which the reply was sent.
 *  seq        unsigned long    The sequence number of the originating
 *                              ping command.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_report_ping_reply(CCBClientLink *cl, CCBLinkType type,
			  unsigned long seq)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(253),
	"ccb_report_ping_reply: NULL argument(s).\n");    
    return 1;
  };
/*
 * If this isn't a reply to the latest ping request, ignore it.
 */
  if(seq != cl->ping_count)
    return 0;
/*
 * Record the fact that a ping-reply was received from the specified link.
 */
  cl->ping_echos |= type;
  return 0;
}

/*.......................................................................
 * Return the set of links over which replies have been received to the
 * most recent ping command.
 *
 * Input:
 *  cl    CCBClientLink *  The networking interface of a CCB client.
 * Output:
 *  return     unsigned    The bitwise union of links that have
 *                         responded. If all is well this should
 *                         be CCB_ALL_LINKS.
 */
unsigned ccb_ping_echos(CCBClientLink *cl)
{
/*
 * Check the arguments.
 */
  if(!cl) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(254),
	"ccb_ping_echos: NULL argument(s).\n");    
    return 0;
  };
/*
 * Return the bitwise union of links over which ping echos have been
 * received.
 */
  return cl->ping_echos;
}

/*.......................................................................
 * Register an optional callback function that will thereafter be used to
 * report changes in the CCB socket file descriptors and the events that
 * they are awaiting.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  fn   CCBClientSocketsFn *  The function to call to report socket and
 *                             socket event changes.
 *  data               void *  Anonymous data to pass to fn().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_client_sockets_callback(CCBClientLink *cl, CCBClientSocketsFn *fn,
				void *data)
{
/*
 * Check the arguments.
 */
  if(!cl || !fn) {
    ccb_log_client_msg(cl, CCB_ERROR, CCB_LOGID(266),
	"ccb_client_sockets_callback: NULL argument(s).\n");
    return 1;
  };
/*
 * Record the new callback, and its data.
 */
  cl->cb.client_sockets_fn = fn;
  cl->cb.client_sockets_data = data;
/*
 * Report the current information. Note that this call is important
 * not only to tell the caller about the current situtation, but also
 * to force-initialize cl->sock_cache.
 */
  if(ccb_report_client_sockets(cl, 1))
    return 1;
  return 0;
}

/*.......................................................................
 * If a suitable callback has been provided, report changes in socket
 * file descriptors and the events that they are awaiting.
 *
 * Input:
 *  cl        CCBClientLink *  The client side of the telemetry and control
 *                             links to the CCB.
 *  force               int    If true, call the application's callback
 *                             even in nothing appears to have changed.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
static int ccb_report_client_sockets(CCBClientLink *cl, int force)
{
/*
 * Do we have a suitable callback function?
 */
  if(cl->cb.client_sockets_fn) {
/*
 * Get the latest socket info.
 */
    int cntrl_sock = ccb_cc_socket(cl->cc);
    int telem_sock = ccb_tc_socket(cl->tc);
    unsigned io = ccb_client_io_status(cl);
/*
 * Unless forced, only call the callback function if anything has changed
 * since it was last called.
 */
    if(force || cntrl_sock != cl->sock_cache.cntrl_sock ||
       telem_sock != cl->sock_cache.telem_sock ||
       io != cl->sock_cache.io) {
/*
 * Use the application's callback function to report the changes.
 */
      if(cl->cb.client_sockets_fn(cl, cl->cb.client_sockets_data,
				 cntrl_sock, telem_sock, io))
	return 1;
/*
 * Update the cache.
 */
      cl->sock_cache.cntrl_sock = cntrl_sock;
      cl->sock_cache.telem_sock = telem_sock;
      cl->sock_cache.io = io;
    };
  };
  return 0;
}
