#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "ccbcommon.h"
#include "ccbconstants.h"
#include "ccbprint.h"

/*
 * Structures of the following type encapsulate a snapshot of the
 * configuration parameters of the CCB.
 */
struct CCBConfig {
  CCBPhaseSwitchCnf phase_switch; /* Phase-switch configuration parameters */
  CCBCalDiodeCnf cal_diode;       /* Cal-diode configuration parameters */
  CCBTimingCnf timing;            /* Timing configuration parameters */
};

/*.......................................................................
 * Create a new CCBConfig object.
 *
 * Output:
 *  return      CCBConfig *  The new object, or NULL on error.
 */
CCBConfig *new_CCBConfig(void)
{
  CCBConfig *cnf;  /* The object to be returned */
/*
 * Allocate the container.
 */
  cnf = malloc(sizeof(CCBConfig));
  if(!cnf) {
    errno = ENOMEM;
    return NULL;
  };
/*
 * Otherwise install the default power-on configuration of the CCB.
 */
  if(ccb_default_config(cnf))
    return del_CCBConfig(cnf);
  return cnf;
}

/*.......................................................................
 * Delete a CCBConfig object.
 *
 * Input:
 *  cnf    CCBConfig *  The object to be deleted.
 * Output:
 *  return CCBConfig *  The deleted object (always NULL).
 */
CCBConfig *del_CCBConfig(CCBConfig *cnf)
{
  if(cnf) {
/*
 * Delete the container.
 */
    free(cnf);
  };
  return NULL;
}

/*.......................................................................
 * Initialize a CCB configuration object with the power-on default
 * configuration of the CCB.
 *
 * Input:
 *  cnf      CCBConfig *  The configuration object to initialize.
 * Output:
 *  return         int    0 - OK.
 *                        1 - Error.
 */
int ccb_default_config(CCBConfig *cnf)
{
  int i;
/*
 * Check the arguments.
 */
  if(!cnf) {
    errno = EINVAL;
    return 1;
  };
/*
 * Install the default power-on phase-switch configuration.
 */ 
  cnf->phase_switch.active_switches = CCB_NO_PHASE_SWITCHES;
  cnf->phase_switch.driven_switches = CCB_NO_PHASE_SWITCHES;
  cnf->phase_switch.closed_switches = CCB_NO_PHASE_SWITCHES;
  cnf->phase_switch.samp_per_state = 1;   /* 1-A/D sample per phase-switch */
                                          /*  state. */
/*
 * Install the default power-on cal-diode configuration.
 */ 
  cnf->cal_diode.ncal = 1;                /* One cal step (diodes off) */
  cnf->cal_diode.driven_diodes = CCB_NO_PHASE_SWITCHES;
  for(i=0; i<CCB_MAX_NCAL; i++) {
    cnf->cal_diode.diode_states[i]=CCB_NO_CAL_DIODES; /* Both diodes off */
    cnf->cal_diode.diode_times[i] = 1;    /* 1 integration per cal-diode */
                                          /*  state. */
  };
/*
 * Install the default power-on timing configuration.
 */ 
  cnf->timing.sample_dt = 250;            /* 250 x 100ns = 25us A/D samples */
  cnf->timing.phase_switch_dt = 10;       /* 10 x 100ns = 1us phase-switch */
                                          /*  delay. */
  cnf->timing.analog_reset_dt = 10;       /* 10 x 100ns = 1us integrator */
                                          /*  resets. */
  cnf->timing.diode_rise_dt = 1000;       /* 1000 x 100ns = 1ms cal-diode */
                                          /*  rise-time. */
  cnf->timing.diode_fall_dt = 1000;       /* 1000 x 100ns = 1ms cal-diode */
                                          /*  fall-time. */
  cnf->timing.integ_period = 40;          /* 40 x 1x25us-sample-per-state */
                                          /*  generates 1ms integrations */
  return 0;
}

/*.......................................................................
 * Make a copy of a CCB configuration object.
 *
 * Input:
 *  orig  const CCBConfig *  The object to be copied.
 *  dest        CCBConfig *  The destination of the copy.
 * Output:
 *  return            int    0 - OK.
 *                           1 - orig or dest was NULL.
 */
int ccb_copy_config(const CCBConfig *orig, CCBConfig *dest)
{
  if(!orig || !dest)
    return 1;
  *dest = *orig;
  return 0;
}

/*.......................................................................
 * Install a new phase-switch configuration within a given CCB
 * configuration object.
 *
 * Input:
 *  cnf                   CCBConfig *  The CCB configuration object to be
 *                                     modified.
 *  active_switches  unsigned short    The set of phase switches that
 *                                     are to be switched, expressed
 *                                     as a bitwise union of
 *                                     CCBPhaseSwitches enumerators.
 *                                     Switches that aren't specified
 *                                     to be switched, are held in
 *                                     the positions indicated by the
 *                                     closed_switches argument.
 *  driven_switches  unsigned short    The set of phase switches whose
 *                                     control lines are to be driven,
 *                                     expressed as a bitwise union of
 *                                     CCBPhaseSwitches enumerators.
 *                                     The remaining lines are placed
 *                                     in a high impedance state.
 *  closed_switches  unsigned short    The set of phase switches that
 *                                     are to be closed at the start
 *                                     of each new phase-switch cycle,
 *                                     expressed as a bitwise union of
 *                                     CCBPhaseSwitches enumerators.
 *  samp_per_state   unsigned short    This parameter configures the number
 *                                     of A/D samples that are integrated
 *                                     between changes in the states of
 *                                     either phase-switch. This must be
 *                                     <= 2 raised to -5 plus the number
 *                                     of switches selected via the
 *                                     'active_switches' argument.
 * Output:
 *  return                      int    0 - OK.
 *                                     1 - Error (see errno).
 */
int ccb_set_phase_switch_cnf(CCBConfig *cnf,
			     unsigned short active_switches,
			     unsigned short driven_switches,
			     unsigned short closed_switches,
			     unsigned short samp_per_state)
{
/*
 * Check the arguments.
 */
  if(!cnf) {
    errno = EINVAL;
    return 1;
  };
/*
 * In order that ccb_diff_config() can bytewise compare the contents
 * of CCBPhaseSwitchCnf members in two CCBConfig objects without worrying
 * about any the random values of any padding bytes, zero out the
 * CCBPhaseSwitchCnf member before assigning the new values.
 */
  memset(&cnf->phase_switch, 0, sizeof(cnf->phase_switch));
/*
 * Record the new configuration.
 */
  cnf->phase_switch.active_switches = active_switches &CCB_ALL_PHASE_SWITCHES;
  cnf->phase_switch.driven_switches = driven_switches &CCB_ALL_PHASE_SWITCHES;
  cnf->phase_switch.closed_switches = closed_switches &CCB_ALL_PHASE_SWITCHES;
  cnf->phase_switch.samp_per_state = samp_per_state;
  return 0;
}

/*.......................................................................
 * Install a new calibration-diode configuration in a given CCB
 * configuration object.
 *
 * Input:
 *  cnf                      CCBConfig *  The CCB configuration object to be
 *                                        modified.
 *  driven_diodes       unsigned short    The 2 least significant bits of
 *                                        this parameter specifies which of
 *                                        the two calibration diode control
 *                                        lines are to be driven. A value of
 *                                        1 means that the corresponding line
 *                                        should be driven, whereas a value
 *                                        of 0 means that it should be placed
 *                                        in a high impedance state.
 *  ncal                unsigned short    The number of steps in the
 *                                        calibration diode state
 *                                        machine. This must be less than or
 *                                        equal to 'CCB_MAX_NCAL'. A value
 *                                        zero can be used if the calibration
 *                                        diodes are to be left turned off
 *                                        throughout the parent scan.
 *  diode_states  const unsigned short *  The first 'ncal' elements of this
 *                                        array specify which of the
 *                                        calibration diodes are to be
 *                                        switched on during the
 *                                        corresponding step of the
 *                                        cal-diode state machine. Each
 *                                        element is a bitwise union of
 *                                        CCBCalDiode enumerators.
 *                                        This argument can be 'NULL' if
 *                                        'ncal' is 0.
 *  diode_times    const unsigned long *  Each of the first 'ncal' elements
 *                                        of this parameter specify for how
 *                                        many integrations the state of the
 *                                        corresponding stage of the
 *                                        calibration diode state machine
 *                                        should be maintained. This argument
 *                                        can be 'NULL' if 'ncal' is 0. Note
 *                                        that steps with zero integrations
 *                                        are simply eliminated.
 * Output:     	 	  	     
 *  return                         int    0 - OK.
 *                                        1 - Error.
 */
int ccb_set_cal_diode_cnf(CCBConfig *cnf,
			  unsigned short driven_diodes,
			  unsigned short ncal,
			  const unsigned short *diode_states,
			  const unsigned long *diode_times)
{
  CCBCalDiodeCnf pars;  /* A local copy of the new configuration parameters. */
  int i;
/*
 * Check the arguments.
 */
  if(!cnf || (ncal > 0 && (!diode_states || !diode_times)) ||
     ncal > CCB_MAX_NCAL) {
    errno = EINVAL;
    return 1;
  };
/*
 * In order that ccb_diff_config() can bytewise compare the contents
 * of CCBCalDiodeCnf members in two CCBConfig objects without worrying
 * about any the random values of any padding bytes, zero out the
 * CCBCalDiodeCnf variable before assigning the new values.
 */
  memset(&pars, 0, sizeof(pars));
/*
 * For the sake of optimization, first install the new parameters in
 * the local 'pars' variable.
 */
  pars.driven_diodes = driven_diodes & CCB_ALL_CAL_DIODES;
/*
 * Record all calibration steps that have a finite number of integrations.
 */
  pars.ncal = 0;
  for(i=0; i<ncal; i++) {
    if(diode_times[i] > 0) {
      pars.diode_states[pars.ncal] = diode_states[i];
      pars.diode_times[pars.ncal] = diode_times[i];
      pars.ncal++;
    };
  };
/*
 * If no calibration steps have been specified, this means that the
 * calibration diodes are to be left switched off. This can be
 * represented as a single calibration step of a single integration,
 * with the diodes turned off.
 */
  if(pars.ncal < 1) {
    pars.ncal = 1;
    pars.diode_states[0] = CCB_NO_CAL_DIODES;
    pars.diode_times[0] = 1;
  };
/*
 * Copy the configuration parameters into the configuration object.
 */
  cnf->cal_diode = pars;
  return 0;
}

/*.......................................................................
 * Install a new timing configuration in a given CCB configuration
 * object.
 *
 * Input:
 *  cnf                   CCBConfig *  The CCB configuration object to be
 *                                     modified.
 *  sample_dt        unsigned short    This refers to the amount of time
 *                                     taken per A/D sample, including the
 *                                     time used to blank phase-switch
 *                                     transitions, but not including the
 *                                     time spent resetting the analog
 *                                     integrators. It is expressed as an
 *                                     integer multiplier of 100ns,
 *                                     and has a nominal value of 250
 *                                     (ie. 25us).
 *  phase_switch_dt  unsigned short    This specifies how much of the
 *                                     sample interval is lost to waiting
 *                                     for the phase switches to settle
 *                                     after phase-switch transitions. It
 *                                     is expressed as an integer
 *                                     multiplier of 100ns.
 *  analog_reset_dt  unsigned short    This specifies how long to wait for
 *                                     the analog integrator to reset
 *                                     before starting to integrate a new
 *                                     sample.  It is expressed as an
 *                                     integer multiplier of 100ns, so
 *                                     for the predicted reset-time of
 *                                     around 1us, 'analog_reset_dt'
 *                                     it would have to be 10.
 *  diode_rise_dt     unsigned long    This specifies the minimum time to
 *                                     wait before starting any integration
 *                                     that starts with either of the
 *                                     calibration diodes being newly
 *                                     switched on. It is expressed as an
 *                                     integer multiplier of 100ns.
 *  diode_fall_dt     unsigned long    This specifies the minimum time to
 *                                     wait before starting any integration
 *                                     that starts with either of the
 *                                     calibration diodes being newly
 *                                     switched off. It is expressed as an
 *                                     integer multiplier of 100ns.
 *  integ_period      unsigned long    This specifies the number of
 *                                     phase-switch cycles that are co-added
 *                                     to form the integrations that are sent
 *                                     to the manager. The physical length of
 *                                     time that this corresponds to depends
 *                                     on the length of an A/D sample and the
 *                                     number of samples per phase-switch
 *                                     cycle.
 * Output:
 *  return                      int    0 - OK.
 *                                     1 - Error.
 */
int ccb_set_timing_cnf(CCBConfig *cnf,
		       unsigned short sample_dt,
		       unsigned short phase_switch_dt,
		       unsigned short analog_reset_dt,
		       unsigned long diode_rise_dt,
		       unsigned long diode_fall_dt,
		       unsigned long integ_period)
{
/*
 * Check the arguments.
 */
  if(!cnf) {
    errno = EINVAL;
    return 1;
  };
/*
 * In order that ccb_diff_config() can bytewise compare the contents
 * of CCBTimingCnf members in two CCBConfig objects without worrying
 * about any the random values of any padding bytes, zero out the
 * CCBTimingCnf member before assigning the new values.
 */
  memset(&cnf->timing, 0, sizeof(cnf->timing));
/*
 * Install the new configuration.
 */
  cnf->timing.sample_dt = sample_dt;
  cnf->timing.phase_switch_dt = phase_switch_dt;
  cnf->timing.analog_reset_dt = analog_reset_dt;
  cnf->timing.diode_rise_dt = diode_rise_dt;
  cnf->timing.diode_fall_dt = diode_fall_dt;
  cnf->timing.integ_period = integ_period;
  return 0;
}

/*.......................................................................
 * Query the phase-switch configuration within a given CCB
 * configuration object.
 *
 * Input:
 *  cnf       const CCBConfig *  The CCB configuration object to be
 *                               queried.
 * Input/Output:
 *  pars    CCBPhaseSwitchCnf *  The phase-switch configuration parameters
 *                               will be assigned to the variable pointed
 *                               to by this argument.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error (see errno).
 */
int ccb_get_phase_switch_cnf(const CCBConfig *cnf, CCBPhaseSwitchCnf *pars)
{
/*
 * Check the arguments.
 */
  if(!cnf || !pars) {
    errno = EINVAL;
    return 1;
  };
/*
 * Return a copy of the phase-switch configuration parameters.
 */
  *pars = cnf->phase_switch;
  return 0;
}

/*.......................................................................
 * Query the calibration-diode configuration in a given CCB
 * configuration object.
 *
 * Input:
 *  cnf    const CCBConfig *  The CCB configuration object to be
 *                            queried.
 * Input/Output:
 *  pars    CCBCalDiodeCnf *  The cal-diode configuration parameters
 *                            will be assigned to the variable pointed
 *                            to by this argument.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_get_cal_diode_cnf(const CCBConfig *cnf, CCBCalDiodeCnf *pars)
{
/*
 * Check the arguments.
 */
  if(!cnf || !pars) {
    errno = EINVAL;
    return 1;
  };
/*
 * Return a copy of the cal-diode configuration parameters.
 */
  *pars = cnf->cal_diode;
  return 0;
}

/*.......................................................................
 * Query the timing configuration in a given CCB configuration object.
 *
 * Input:
 *  cnf    const CCBConfig *  The CCB configuration object to be queried.
 * Input/Output:
 *  pars      CCBTimingCnf *  The timing configuration parameters
 *                            will be assigned to the variable pointed
 *                            to by this argument.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_get_timing_cnf(const CCBConfig *cnf, CCBTimingCnf *pars)
{
/*
 * Check the arguments.
 */
  if(!cnf || !pars) {
    errno = EINVAL;
    return 1;
  };
/*
 * Return a copy of the timing configuration parameters.
 */
  *pars = cnf->timing;
  return 0;
}

/*.......................................................................
 * Verify that the contents of a CCB configuration object make sense.
 *
 * Note that it is assumed that the contents of *cnf were installed by
 * the various ccb_set_*_cnf() functions above, so the checks that those
 * functions perform are not repeated.
 *
 * Input:
 *  cnf    CCBConfig *   The configuration object to be validated.
 *  errdim    size_t     The allocated dimension of the errmsg[] array.
 * Input/Output:
 *  errmsg      char *   The array of errdim elements in which to
 *                       report any configuration errors.
 * Output:
 *  return       int     0 - OK.
 *                       1 - Invalid configuration. The contents of
 *                           errmsg[] describe the error.
 */
int ccb_check_config(CCBConfig *cnf, size_t errdim, char *errmsg)
{
  CCBInterval dt;  /* The integration period */
/*
 * If no error message buffer has been provided, force errdim to be
 * zero, so that ccb_snprintf() doesn't to try to write to it.
 */
  if(!errmsg)
    errdim = 0;
/*
 * Check the remaining arguments.
 */
  if(!cnf) {
    ccb_snprintf(errmsg, errdim, "NULL argument(s).\n");
    return 1;
  };
/*
 * Check that the minimum A/D readout time can be accomodated.
 */
  if(cnf->timing.phase_switch_dt >= cnf->timing.sample_dt ||
     cnf->timing.sample_dt - cnf->timing.phase_switch_dt <
     CCB_MIN_ADC_READOUT_DT) {
    ccb_snprintf(errmsg, errdim, "ADC sampling time too short");
    return 1;
  };
/*
 * Check the number of samples per phase-switch state.
 */
  if(cnf->phase_switch.samp_per_state == 0) {
    ccb_snprintf(errmsg, errdim, "Zero ADC samples per phase-switch state");
    return 1;
  };
  if(ccb_cycle_length(cnf->phase_switch.active_switches) *
     cnf->phase_switch.samp_per_state > 32) {
    ccb_snprintf(errmsg, errdim,"Too many ADC samples per phase-switch state");
    return 1;
  };
/*
 * Compute the integration period of .
 */
  if(ccb_integration_duration(cnf, 1, &dt)) {
    switch(errno) {
    case ERANGE:
      ccb_snprintf(errmsg, errdim, "Integration period too long to calculate");
      break;
    case EDOM:
      ccb_snprintf(errmsg, errdim,
		   "Divide by zero while computing integration period");
      break;
    default:
      ccb_snprintf(errmsg, errdim, "Error computing integration period");
      break;
    };
    return 1;
  };
/*
 * Check that the integration period isn't too small.
 */
  if(dt.sec==0 && dt.ns < CCB_MIN_INTEG_DT) {
    ccb_snprintf(errmsg, errdim, "Integration period too short.\n");
    return 1;
  };
/*
 * Configuration okay.
 */
  return 0;
}

/*.......................................................................
 * Report the set of configuration groups that differ between two CCB
 * configuration object.
 *
 * Input:
 *  ca    const CCBConfig *  The first of the configurations to be compared.
 *  cb    const CCBConfig *  The second of the configurations to be compared.
 * Output:
 *  return       unsigned    A bitmask union of CCBConfigType enumerators,
 *                           indicating which parts of configuration differ
 *                           between ca and cb.
 */
unsigned ccb_diff_config(const CCBConfig *ca, const CCBConfig *cb)
{
  unsigned mask=0;  /* The bitmask set of differences between ca and cb */
/*
 * If one or both of the configuration objects is NULL, report all
 * configuration parameters as differing.
 */
  if(!ca || !cb)
    return CCB_CNF_ALL;
/*
 * Compare the phase-switch configuration parameters.
 */
  if(memcmp(&ca->phase_switch, &cb->phase_switch, sizeof(ca->phase_switch)))
    mask |= CCB_CNF_PHASE_SWITCHES;
/*
 * Compare the calibration diode configuration parameters.
 */
  if(memcmp(&ca->cal_diode, &cb->cal_diode, sizeof(ca->cal_diode)))
    mask |= CCB_CNF_CAL_DIODES;
/*
 * Compare the timing configuration parameters.
 */
  if(memcmp(&ca->timing, &cb->timing, sizeof(ca->timing)))
    mask |= CCB_CNF_TIMING;
  return mask;
}

/*.......................................................................
 * Set a CCBConfig object to a given configuration.
 *
 * Input:
 *  cnf                   CCBConfig *  The CCB configuration object to be
 *                                     modified.
 *  phase   const CCBPhaseSwitchCnf *  The phase-switch configuration
 *                                     to assign, or NULL to leave this
 *                                     part of the configuration unchanged.
 *  cal        const CCBCalDiodeCnf *  The cal-diode configuration
 *                                     to assign, or NULL to leave this
 *                                     part of the configuration unchanged.
 *  timing       const CCBTimingCnf *  The timing configuration
 *                                     to assign, or NULL to leave this
 *                                     part of the configuration unchanged.
 * Output:
 *  return                      int    0 - OK.
 *                                     1 - Error.
 */
int ccb_set_config(CCBConfig *cnf, const CCBPhaseSwitchCnf *phase,
		   const CCBCalDiodeCnf *cal, const CCBTimingCnf *timing)
{
/*
 * Check the arguments.
 */
  if(!cnf) {
    errno = EINVAL;
    return 1;
  };
/*
 * Validate and install the specified phase-switch configuration.
 */
  if(phase && ccb_set_phase_switch_cnf(cnf, phase->active_switches,
			      phase->driven_switches, phase->closed_switches,
			      phase->samp_per_state))
    return 1;
/*
 * Validate and install the specified cal-diode configuration.
 */
  if(cal && ccb_set_cal_diode_cnf(cnf, cal->driven_diodes, cal->ncal,
				  cal->diode_states, cal->diode_times))
    return 1;
/*
 * Validate and install the specified timing configuration.
 */
  if(timing && ccb_set_timing_cnf(cnf, timing->sample_dt,
		  timing->phase_switch_dt, timing->analog_reset_dt,
		  timing->diode_rise_dt, timing->diode_fall_dt,
                  timing->integ_period))
    return 1;
  return 0;
}
