#ifndef ccbcontrolmsg_h
#define ccbcontrolmsg_h

#include "netbuf.h"
#include "ccbnetobj.h"

/*-----------------------------------------------------------------------*
 * This header defines the message data-structures and description       *
 * tables that are used to exchange control commands and replies         *
 * between a CCB manager and a CCB server. Note that although this is    *
 * a header file, the message-description table and its contents are     *
 * all declared static. This prevents namespace polution while keeping   *
 * all definitions in one place. An alternative would be to place the    *
 * message data-structures in this header, but the corresponding         *
 * message-table description entries in a separate C file. This is       *
 * judged to be a recipe for errors, and since this header file is only  *
 * included by 2 C files, ccbcontrolserver.c and ccbcontrolclient.c,     *
 * and the table is trivially small, the overhead of defining the table  *
 * anew wherever this header is included is judged to be insignificant.  *
 *-----------------------------------------------------------------------*/

#ifndef CCB_ARRAY_DIM
#define CCB_ARRAY_DIM(array) (sizeof(array)/sizeof(array[0]))
#endif

/*-----------------------------------------------------------------------
 * Separately define each of the command message structures used to send
 * messages from the manager to the CCB server.
 *-----------------------------------------------------------------------*/

/*
 * Define the enumerators and structures of command messages sent by the
 * CCB manager to the CCB server.
 */
typedef enum {
  CCB_PHASE_SWITCH_CMD,  /* A phase-switch config command */
  CCB_CAL_DIODE_CMD,     /* A cal-diode config command */
  CCB_TIMING_CMD,        /* An timing config command */
  CCB_START_SCAN_CMD,    /* A start-scan command */
  CCB_STOP_SCAN_CMD,     /* A stop-scan command */
  CCB_MONITOR_CMD,       /* A monitoring control command */
  CCB_TELEMETRY_CMD,     /* A telemetry stream control command */
  CCB_LOGGER_CMD,        /* A log control command */
  CCB_RESET_CMD,         /* A reset command */
  CCB_AWAKEN_CMD,        /* An awaken command */
  CCB_STANDBY_CMD,       /* A standby command */
  CCB_PING_CMD,          /* a ping command */
  CCB_STATUS_REQUEST_CMD,/* a check-status command */
  CCB_SHUTDOWN_CMD,      /* A computer shutdown command */
  CCB_REBOOT_CMD,        /* A computer reboot command */
} CCBControlCommandType;

/*
 * A member of the following type is the first member of all control
 * messages.
 */
typedef struct {
  CCBNetMsg base;   /* The initial member of all messages */
  long id;          /* The manager's identifier of the message */
} CCBControlCommandHeader;

/*
 * Provide a macro to describe the initial member of all control
 * command message structures. Note that the base member is omitted,
 * since this member is already known of by the ccbnetobj module.
 */
#define CCB_CONTROL_CMD_HEADER \
  {"base",  offsetof(CCBControlCommandHeader,base), CCB_NET_MSG,  1}, \
  {"id",    offsetof(CCBControlCommandHeader,id),   CCB_NET_LONG, 1},

/* A phase-switch configuration command */

typedef struct {
  CCBControlCommandHeader head; /* head.base.type=CCB_PHASE_SWITCH_CMD */
  CCBPhaseSwitchCnf cnf;        /* The phase-switch configuration parameters */
} CCBPhaseSwitchCmd;

static const CCBNetMsgMember ccb_phase_switch_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"active_switches", offsetof(CCBPhaseSwitchCmd,cnf.active_switches),
     CCB_NET_USHORT, 1},
  {"driven_switches", offsetof(CCBPhaseSwitchCmd,cnf.driven_switches),
     CCB_NET_USHORT,  1},
  {"closed_switches", offsetof(CCBPhaseSwitchCmd,cnf.closed_switches),
     CCB_NET_USHORT, 1},
  {"samp_per_state",  offsetof(CCBPhaseSwitchCmd,cnf.samp_per_state),
     CCB_NET_USHORT, 1},
};

/* A calibration-diode configuration command */

typedef struct {
  CCBControlCommandHeader head;    /* head.base.type=CCB_CAL_DIODE_CMD*/
  CCBCalDiodeCnf cnf;              /* The cal-diode configuration parameters */
} CCBCalDiodeCmd;

static const CCBNetMsgMember ccb_cal_diode_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"driven_diodes",   offsetof(CCBCalDiodeCmd,cnf.driven_diodes),
     CCB_NET_USHORT, 1},
  {"ncal",            offsetof(CCBCalDiodeCmd,cnf.ncal),
     CCB_NET_USHORT, 1},
  {"diode_states",    offsetof(CCBCalDiodeCmd,cnf.diode_states),
     CCB_NET_USHORT, CCB_MAX_NCAL},
  {"diode_times",     offsetof(CCBCalDiodeCmd,cnf.diode_times),
     CCB_NET_ULONG, CCB_MAX_NCAL},
};

/* An acquisition-timing configuration command */

typedef struct {
  CCBControlCommandHeader head;  /* head.base.type=CCB_TIMING_CMD */
  CCBTimingCnf cnf;              /* The timing configuration parameters */
} CCBTimingCmd;

static const CCBNetMsgMember ccb_timing_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"sample_dt",       offsetof(CCBTimingCmd,cnf.sample_dt),
     CCB_NET_USHORT, 1},
  {"phase_switch_dt", offsetof(CCBTimingCmd,cnf.phase_switch_dt),
     CCB_NET_USHORT, 1},
  {"analog_reset_dt", offsetof(CCBTimingCmd,cnf.analog_reset_dt),
     CCB_NET_USHORT, 1},
  {"diode_rise_dt",   offsetof(CCBTimingCmd,cnf.diode_rise_dt),
     CCB_NET_ULONG, 1},
  {"diode_fall_dt",   offsetof(CCBTimingCmd,cnf.diode_fall_dt),
     CCB_NET_ULONG, 1},
  {"integ_period",    offsetof(CCBTimingCmd,cnf.integ_period),
     CCB_NET_ULONG, 1},
};

/* A start-scan command */

typedef struct {
  CCBControlCommandHeader head;  /* head.base.type=CCB_START_SCAN_CMD */
  unsigned long scan;            /* The numeric ID to give the new scan */
  unsigned long mjd;             /* The MJD UTC day number */
  unsigned long tod;             /* The time of day (seconds since */
                                 /*  0H UTC). */
} CCBStartScanCmd;

static const CCBNetMsgMember ccb_start_scan_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"scan",   offsetof(CCBStartScanCmd,scan),   CCB_NET_ULONG, 1},
  {"mjd",    offsetof(CCBStartScanCmd,mjd),    CCB_NET_ULONG, 1},
  {"tod",    offsetof(CCBStartScanCmd,tod),    CCB_NET_ULONG, 1},
};

/* A stop-scan command */

typedef struct {
  CCBControlCommandHeader head;  /* head.base.type=CCB_STOP_SCAN_CMD */
  unsigned long scan;            /* The numeric ID to give the new intra-scan */
} CCBStopScanCmd;

static const CCBNetMsgMember ccb_stop_scan_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"scan",   offsetof(CCBStopScanCmd,scan),   CCB_NET_ULONG, 1},
};

/* A reset command */

typedef struct {
  CCBControlCommandHeader head;     /* head.base.type=CCB_RESET_CMD */
} CCBResetCmd;

static const CCBNetMsgMember ccb_reset_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
};

/* A standby command */

typedef struct {
  CCBControlCommandHeader head;  /* head.base.type=CCB_STANDBY_CMD */
  unsigned short stream_mask;      /* The telemetry selection mask */
} CCBStandbyCmd;

static const CCBNetMsgMember ccb_standby_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"stream_mask", offsetof(CCBStandbyCmd,stream_mask), CCB_NET_USHORT, 1},
};

/* An awaken command */

typedef struct {
  CCBControlCommandHeader head;   /* head.base.type=CCB_AWAKEN_CMD */
} CCBAwakenCmd;

static const CCBNetMsgMember ccb_awaken_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
};

/* A ping command */

typedef struct {
  CCBControlCommandHeader head;   /* head.base.type=CCB_PING_CMD */
  unsigned long seq;              /* The sequence number of the ping. */
} CCBPingCmd;

static const CCBNetMsgMember ccb_ping_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"seq",      offsetof(CCBPingCmd,seq), CCB_NET_ULONG,  1},
};

/* A status-request command */

typedef struct {
  CCBControlCommandHeader head;    /* head.base.type=CCB_STATUS_REQUEST_CMD */
} CCBStatusRequestCmd;

static const CCBNetMsgMember ccb_status_request_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
};

/* A shutdown command */

typedef struct {
  CCBControlCommandHeader head;   /* head.base.type=CCB_SHUTDOWN_CMD */
} CCBShutdownCmd;

static const CCBNetMsgMember ccb_shutdown_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
};

/* A reboot command */

typedef struct {
  CCBControlCommandHeader head;    /* head.base.type=CCB_REBOOT_CMD */
} CCBRebootCmd;

static const CCBNetMsgMember ccb_reboot_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
};

/* A monitoring control-command */

typedef struct {
  CCBControlCommandHeader head;    /* head.base.type=CCB_MONITOR_CMD */
  unsigned short period;           /* The interval between monitoring */
                                   /*  updates. */
} CCBMonitorCmd;

static const CCBNetMsgMember ccb_monitor_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"period",    offsetof(CCBMonitorCmd,period), CCB_NET_USHORT, 1},
};

/* A telemetry-stream control-command */

typedef struct {
  CCBControlCommandHeader head;    /* head.base.type=CCB_TELEMETRY_CMD */
  unsigned short streams;          /* The telemetry-streams to report */
} CCBTelemetryCmd;

static const CCBNetMsgMember ccb_telemetry_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"streams",   offsetof(CCBTelemetryCmd,streams), CCB_NET_USHORT, 1},
};

/* A log control-command */

typedef struct {
  CCBControlCommandHeader head;  /* head.base.type=CCB_LOGGER_CMD */
  unsigned long period;          /* The interval at which historical */
                                 /*  log messages are forgotten (seconds). */
} CCBLoggerCmd;

static const CCBNetMsgMember ccb_logger_cmd_members[] = {
  CCB_CONTROL_CMD_HEADER
  {"period",    offsetof(CCBLoggerCmd,period), CCB_NET_ULONG, 1},
};

/*
 * Create a union of the above message containers.
 */
typedef union {
  CCBControlCommandHeader head;/* The common control message header */
  CCBPhaseSwitchCmd phase_cmd; /* head.base.type=CCB_PHASE_SWITCH_CMD */
  CCBCalDiodeCmd diode_cmd;    /* head.base.type=CCB_CAL_DIODE_CMD */
  CCBTimingCmd timing_cmd;     /* head.base.type=CCB_TIMING_CMD */
  CCBStartScanCmd start_scan;  /* head.base.type=CCB_START_SCAN_CMD */
  CCBStopScanCmd stop_scan;    /* head.base.type=CCB_STOP_SCAN_CMD */
  CCBMonitorCmd monitor;       /* head.base.type=CCB_MONITOR_CMD */
  CCBTelemetryCmd telemetry;   /* head.base.type=CCB_TELEMETRY_CMD */
  CCBLoggerCmd logger;         /* head.base.type=CCB_LOGGER_CMD */
  CCBResetCmd reset;           /* head.base.type=CCB_RESET_CMD */
  CCBAwakenCmd awaken;         /* head.base.type=CCB_AWAKEN_CMD */
  CCBStandbyCmd standby;       /* head.base.type=CCB_STANDBY_CMD */
  CCBPingCmd ping;             /* head.base.type=CCB_PING_CMD */
  CCBStatusRequestCmd status;  /* head.base.type=CCB_STATUS_REQUEST_CMD */
  CCBShutdownCmd shutdown;     /* head.base.type=CCB_SHUTDOWN_CMD */
  CCBRebootCmd reboot;         /* head.base.type=CCB_REBOOT_CMD */
} CCBControlCommand;

/*
 * Collect all of the object definitions in an array.
 */
static const CCBNetMsgInfo ccb_control_cmd_structures[] = {
  {CCB_PHASE_SWITCH_CMD,      "phase_switch_cmd",
   ccb_phase_switch_cmd_members, CCB_ARRAY_DIM(ccb_phase_switch_cmd_members),
   sizeof(CCBPhaseSwitchCmd)},
  {CCB_CAL_DIODE_CMD,      "cal_diode_cmd",
   ccb_cal_diode_cmd_members, CCB_ARRAY_DIM(ccb_cal_diode_cmd_members),
   sizeof(CCBCalDiodeCmd)},
  {CCB_TIMING_CMD,      "timing_cmd",
   ccb_timing_cmd_members, CCB_ARRAY_DIM(ccb_timing_cmd_members),
   sizeof(CCBTimingCmd)},
  {CCB_START_SCAN_CMD,      "start_scan_cmd",
   ccb_start_scan_cmd_members, CCB_ARRAY_DIM(ccb_start_scan_cmd_members),
   sizeof(CCBStartScanCmd)},
  {CCB_STOP_SCAN_CMD,      "stop_scan_cmd",
   ccb_stop_scan_cmd_members, CCB_ARRAY_DIM(ccb_stop_scan_cmd_members),
   sizeof(CCBStopScanCmd)},
  {CCB_MONITOR_CMD,      "monitor_cmd",
   ccb_monitor_cmd_members, CCB_ARRAY_DIM(ccb_monitor_cmd_members),
   sizeof(CCBMonitorCmd)},
  {CCB_TELEMETRY_CMD,      "telemetry_cmd",
   ccb_telemetry_cmd_members, CCB_ARRAY_DIM(ccb_telemetry_cmd_members),
   sizeof(CCBTelemetryCmd)},
  {CCB_LOGGER_CMD,      "logger_cmd",
   ccb_logger_cmd_members, CCB_ARRAY_DIM(ccb_logger_cmd_members),
   sizeof(CCBLoggerCmd)},
  {CCB_RESET_CMD,      "reset_cmd",
   ccb_reset_cmd_members, CCB_ARRAY_DIM(ccb_reset_cmd_members),
   sizeof(CCBResetCmd)},
  {CCB_AWAKEN_CMD,      "awaken_cmd",
   ccb_awaken_cmd_members, CCB_ARRAY_DIM(ccb_awaken_cmd_members),
   sizeof(CCBAwakenCmd)},
  {CCB_STANDBY_CMD,      "standby_cmd",
   ccb_standby_cmd_members, CCB_ARRAY_DIM(ccb_standby_cmd_members),
   sizeof(CCBStandbyCmd)},
  {CCB_PING_CMD,      "ping_cmd",
   ccb_ping_cmd_members, CCB_ARRAY_DIM(ccb_ping_cmd_members),
   sizeof(CCBPingCmd)},
  {CCB_STATUS_REQUEST_CMD,      "status_request_cmd",
   ccb_status_request_cmd_members,
     CCB_ARRAY_DIM(ccb_status_request_cmd_members),
     sizeof(CCBStatusRequestCmd)},
  {CCB_SHUTDOWN_CMD,      "shutdown_cmd",
   ccb_shutdown_cmd_members, CCB_ARRAY_DIM(ccb_shutdown_cmd_members),
   sizeof(CCBShutdownCmd)},
  {CCB_REBOOT_CMD,      "reboot_cmd",
   ccb_reboot_cmd_members, CCB_ARRAY_DIM(ccb_reboot_cmd_members),
   sizeof(CCBRebootCmd)},
};

/*
 * Form the table of objects that is used by the functions in
 * netobj.h.
 */
static const CCBNetMsgTable ccb_control_cmd_table = {
  "control_command", ccb_control_cmd_structures,
  sizeof(ccb_control_cmd_structures)/sizeof(CCBNetMsgInfo),
  sizeof(CCBControlCommand)
};

/*-----------------------------------------------------------------------
 * Describe each of the message structures used by the CCB server to
 * send replies to the manager.
 *-----------------------------------------------------------------------*/

/*
 * Define the enumerators and structures of reply messages sent by the
 * CCB server to the CCB manager.
 */
typedef enum {
  CCB_CNTRL_PING_REPLY,   /* A ping reply */
  CCB_STATUS_REPLY, /* A check-status reply */
  CCB_CNTRL_CMD_ACK       /* A control command-acknowledgement reply */
} CCBControlReplyType;

/*
 * The first member of all control-reply message structures is a
 * CCBControlReplyHeader structure.
 */
typedef struct {
  CCBNetMsg base;         /* The base-class of all messages */
} CCBControlReplyHeader;

/*
 * Provide a macro to describe the initial member of all control
 * reply message structures. Currently there aren't any, except for
 * the obligatory base member of all message structures.
 */
#define CCB_CONTROL_REPLY_HEADER \
  {"base", offsetof(CCBControlReplyHeader,base), CCB_NET_MSG, 1},

/* A reply to a ping command */

typedef struct {
  CCBControlReplyHeader head;    /* head.base.type=CCB_CNTRL_PING_REPLY */
  unsigned long seq;             /* The sequence number of the ping */
                                 /*  that is being replied to. */
} CCBCntrlPingReply;

static const CCBNetMsgMember ccb_cntrl_ping_reply_members[] = {
  CCB_CONTROL_REPLY_HEADER
  {"seq",      offsetof(CCBCntrlPingReply,seq), CCB_NET_ULONG,  1},
};

/* A reply to a check-status command */

typedef struct {
  CCBControlReplyHeader head;    /* head.base.type=CCB_STATUS_REPLY */
  unsigned long status;          /* The status of the CCB */
} CCBStatusReply;

static const CCBNetMsgMember ccb_status_reply_members[] = {
  CCB_CONTROL_REPLY_HEADER
  {"status",      offsetof(CCBStatusReply,status),   CCB_NET_ULONG,  1},
};

/* A command acknowledgment reply */

typedef struct {
  CCBControlReplyHeader head;   /* head.base.type=CCB_CNTRL_CMD_ACK */
  unsigned long id;             /* The manager-specified id of the */
                                /*  command being acknowledged */
  unsigned long status;         /* A CCBCmdStatus enumerator */
} CCBCntrlCmdAck;

static const CCBNetMsgMember ccb_cntrl_cmd_ack_members[] = {
  CCB_CONTROL_REPLY_HEADER
  {"id",          offsetof(CCBCntrlCmdAck,id),       CCB_NET_ULONG,  1},
  {"status",      offsetof(CCBCntrlCmdAck,status),   CCB_NET_ULONG,  1},
};

/*
 * Create a union of the above message containers.
 */
typedef union {
  CCBControlReplyHeader head;   /* The common header of all control */
                                /*  link replies. */
  CCBCntrlPingReply ping;        /* base.type = CCB_CNTRL_PING_REPLY */
  CCBStatusReply status;    /* base.type = CCB_STATUS_REPLY */
  CCBCntrlCmdAck cmd_ack;        /* base.type = CCB_CNTRL_CMD_ACK */
} CCBControlReply;

/*
 * Collect all of the object definitions in an array.
 */
static const CCBNetMsgInfo ccb_control_reply_structures[] = {
  {CCB_CNTRL_PING_REPLY,        "cntrl_ping_reply",
   ccb_cntrl_ping_reply_members, CCB_ARRAY_DIM(ccb_cntrl_ping_reply_members),
   sizeof(CCBCntrlPingReply)},
  {CCB_STATUS_REPLY,      "status_reply",
   ccb_status_reply_members, CCB_ARRAY_DIM(ccb_status_reply_members),
   sizeof(CCBStatusReply)},
  {CCB_CNTRL_CMD_ACK,           "cntrl_cmd_ack",
   ccb_cntrl_cmd_ack_members, CCB_ARRAY_DIM(ccb_cntrl_cmd_ack_members),
   sizeof(CCBCntrlCmdAck)},
};

/*
 * Form the table of objects that is used by the functions in
 * netobj.h.
 */
static const CCBNetMsgTable ccb_control_reply_table = {
  "control_reply", ccb_control_reply_structures,
  sizeof(ccb_control_reply_structures)/sizeof(CCBNetMsgInfo),
  sizeof(CCBControlReply)
};

#endif
