#include <stdlib.h>
#include <limits.h>
#include <errno.h>

#include "ccbcommon.h"
#include "ccbconstants.h"
#include "ccbunsigned.h"

/*
 * Parameterize the number of long integers needed to store the number
 * of nanoseconds stored in a CCBInterval.
 */
#define LONG_PER_INTERVAL 2

/*
 * When an interval is converted from (seconds,nanoseconds) to
 * nanoseconds the bits of the latter large unsigned integer is stored
 * as an array of two unsigned integers.
 */
typedef unsigned long CCBNanoseconds[LONG_PER_INTERVAL];

/*
 * Convert from a CCBInterval value to a CCBNanoseconds value.
 */
static int ccb_interval_to_ns(const CCBInterval *dt, CCBNanoseconds ns);

/*
 * Convert from a CCBNanoseconds value to a CCBInterval value.
 */
static int ccb_ns_to_interval(const CCBNanoseconds ns, CCBInterval *dt);

/*.......................................................................
 * Scale a time interval by an unsigned long integer.
 *
 * Input:
 *  dt   const CCBInterval *  The time interval to be scaled.
 *  factor   unsigned long    The scale factor to apply.
 * Input/Output:
 *  ans        CCBInterval *  The scaled time interval (note that
 *                            'ans' and 'dt' can point to the same
 *                            variable).
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error (see errno).
 */
int ccb_scale_interval(CCBInterval *dt, unsigned long factor, CCBInterval *ans)
{
  CCBNanoseconds dt_ns;      /* The total number of nanoseconds in 'dt' */
  CCBNanoseconds scaled_ns;  /* dt_ns * factor */
  CCBNanoseconds work;       /* A work array */
/*
 * Check the arguments.
 */
  if(!dt || !ans) {
    errno = EINVAL;
    return 1;
  };
/*
 * Convert the interval into an extended integer that represents the
 * total number of nanoseconds in the interval.
 */
  if(ccb_interval_to_ns(dt, dt_ns))
    return 1;
/*
 * Scale the interval by the requested amount.
 */
  if(ccb_ull_multiply(dt_ns, 2, &factor, 1, scaled_ns, 2, work, 2))
    return 1;
/*
 * Convert back from nanoseconds to seconds and nanoseconds.
 */
  if(ccb_ns_to_interval(scaled_ns, ans))
    return 1;
  return 0;
}

/*.......................................................................
 * Add two time intervals.
 *
 * Input:
 *  dt1  const CCBInterval *  The first of the time intervals to added.
 *  dt2  const CCBInterval *  The second of the time intervals to be added.
 * Input/Output:
 *  sum        CCBInterval *  The sum of the intervals. Note that (note
 *                            that 'sum' and 'dt1' can point to the same
 *                            variable).
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error (see errno).
 */
int ccb_add_intervals(CCBInterval *dt1, const CCBInterval *dt2,
		      CCBInterval *sum)
{
  CCBNanoseconds ns1;    /* The total number of nanoseconds in dt1 */
  CCBNanoseconds ns2;    /* The total number of nanoseconds in dt2 */
  CCBNanoseconds ns_sum; /* The sum of ns1 and ns2 */
/*
 * Check the arguments.
 */
  if(!dt1 || !dt2 || !sum) {
    errno = EINVAL;
    return 1;
  };
/*
 * Convert the intervals to extended integers that represent the
 * total number of nanoseconds in each interval.
 */
  if(ccb_interval_to_ns(dt1, ns1) ||
     ccb_interval_to_ns(dt2, ns2))
    return 1;
/*
 * Add the two intervals.
 */
  if(ccb_ull_add(ns1, LONG_PER_INTERVAL, ns2, LONG_PER_INTERVAL,
		 ns_sum, LONG_PER_INTERVAL))
    return 1;
/*
 * Conver the sum back to an interval.
 */
  if(ccb_ns_to_interval(ns_sum, sum))
    return 1;
  return 0;
}

/*.......................................................................
 * Subtract a small time interval from a larger time interval.
 *
 * Input:
 *  dt1  const CCBInterval *  The larger time-interval.
 *  dt2  const CCBInterval *  The time interval to subtract from dt1.
 * Input/Output:
 *  dif        CCBInterval *  The difference dt1-dt2. Note that (note
 *                            that 'dif' and 'dt1' can point to the same
 *                            variable).
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error (see errno).
 */
int ccb_subtract_interval(CCBInterval *dt1, CCBInterval *dt2,
			  CCBInterval *dif)
{
  CCBNanoseconds ns1;     /* The total number of nanoseconds in dt1 */
  CCBNanoseconds ns2;     /* The total number of nanoseconds in dt2 */
  CCBNanoseconds ns_dif;  /* The difference ns1-ns2 */
/*
 * Check the arguments.
 */
  if(!dt1 || !dt2 || !dif) {
    errno = EINVAL;
    return 1;
  };
/*
 * Convert the intervals to extended integers that represent the
 * total number of nanoseconds in each interval.
 */
  if(ccb_interval_to_ns(dt1, ns1) ||
     ccb_interval_to_ns(dt2, ns2))
    return 1;
/*
 * Subtract ns2 from ns1.
 */
  if(ccb_ull_subtract(ns1, LONG_PER_INTERVAL, ns2, LONG_PER_INTERVAL,
		 ns_dif, LONG_PER_INTERVAL))
    return 1;
/*
 * Conver the difference back to an interval.
 */
  if(ccb_ns_to_interval(ns_dif, dif))
    return 1;
  return 0;
}

/*.......................................................................
 * Compare two time intervals.
 *
 * Input:
 *  dt1  const CCBInterval *  The leftmost term of the comparison.
 *  dt2  const CCBInterval *  The rightmost term of the comparison.
 * Output:
 *  return             int    -1:  dt1 < dt2
 *                             0:  dt1 == dt2
 *                             1:  dt1 > dt2
 */
int ccb_compare_intervals(const CCBInterval *dt1, const CCBInterval *dt2)
{
/*
 * Compare the seconds part of the interval, this being the most significant
 * part.
 */
  if(dt1->sec > dt2->sec)
    return 1;
  else if(dt1->sec < dt2->sec)
    return -1;
/*
 * The two seconds terms are equal, so compare the nanoseconds parts.
 */
  else if(dt1->ns > dt2->ns)
    return 1;
  else if(dt1->ns < dt2->ns)
    return -1;
/*
 * Both the seconds and the nano-seconds terms are equal.
 */
  return 0;
}

/*.......................................................................
 * Initialize an interval to 0.
 *
 * Input/Output:
 *  dt     CCBInterval *  The interval to be zeroed.
 */
void ccb_zero_interval(CCBInterval *dt)
{
  if(dt) {
    dt->sec = 0;
    dt->ns = 0;
  };
}

/*.......................................................................
 * Convert a duration specified as a number of hardware clock ticks to
 * a CCBInterval.
 *
 * Input:
 *  ticks    unsigned long    The duration as a number of clock ticks.
 * Input/Output:
 *  dt         CCBInterval *  The corresponding time interval will be
 *                            assigned to *dt.
 */
void ccb_clock_interval(unsigned long ticks, CCBInterval *dt)
{
/*
 * Split the clock time into seconds and nanoseconds.
 */
  if(dt) {
    dt->sec = ticks / CCB_CLOCK_FREQ;
    dt->ns = (ticks - dt->sec * CCB_CLOCK_FREQ) * CCB_NS_PER_TICK;
  };
}

/*.......................................................................
 * Convert an interval from its split seconds,nanosecond representation
 * to a total number of nanoseconds.
 *
 * Input:
 *  dt  const CCBInterval *  The interval to be converted.
 * Input/Output:
 *  ns     CCBNanoseconds    The number of nanoseconds corresponding
 *                           to the above interval, will be stored in
 *                           ns[].
 * Output:
 *  return            int    0 - OK.
 *                           1 - Error (see errno).
 */
static int ccb_interval_to_ns(const CCBInterval *dt, CCBNanoseconds ns)
{
  CCBNanoseconds work;  /* A work arrays */
/*
 * Get the number of nanoseconds per second.
 */
  const unsigned long ns_per_sec = CCB_NS_PER_SEC;
/*
 * Check the arguments.
 */
  if(!dt || !ns) {
    errno = EINVAL;
    return 1;
  };
/*
 * Scale the seconds component of 'ns', by 1e9, to convert it to
 * nanoseconds, and assign the result to 'ns'.
 */
  if(ccb_ull_multiply(&dt->sec, 1, &ns_per_sec, 1, ns, LONG_PER_INTERVAL,
		      work, LONG_PER_INTERVAL))
    return 1;
/*
 * Add the nanosecond component of 'dt' to 'ns'.
 */
  if(ccb_ull_add(ns, LONG_PER_INTERVAL, &dt->ns, 1, ns, LONG_PER_INTERVAL))
    return 1;
  return 0;
}

/*.......................................................................
 * Convert a total number of nanoseconds to a CCBInterval.
 *
 * Input:
 *  ns  const CCBNanoseconds    The nanoseconds count to be converted.
 *                              Note that zero nano-seconds can be
 *                              expressed by passing 'ns' as NULL.
 * Input/Output:
 *  dt           CCBInterval *  The CCBInterval version of 'ns'.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int ccb_ns_to_interval(const CCBNanoseconds ns, CCBInterval *dt)
{
  CCBNanoseconds sec;  /* The seconds component of the interval */
  CCBNanoseconds nsec; /* The nano-seconds component of the interval */
/*
 * Get the number of nanoseconds per second.
 */
  const unsigned long ns_per_sec = CCB_NS_PER_SEC;
/*
 * Check the arguments.
 */
  if(!dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Divide the interval by the number of nanoseconds in an
 * integer. Since this will be an integer division, the resulting
 * quotient will be the number of whole seconds in the interval, and
 * the remainder will be the remaining number of nanoseconds.
 */
  if(ccb_ull_divide(ns, LONG_PER_INTERVAL, &ns_per_sec, 1,
		    sec, LONG_PER_INTERVAL, nsec, LONG_PER_INTERVAL))
    return 1;
/*
 * Assign the return values.
 */
  if(ccb_ull_copy(sec, LONG_PER_INTERVAL, &dt->sec, 1) ||
     ccb_ull_copy(nsec, LONG_PER_INTERVAL, &dt->ns, 1))
    return 1;
  return 0;
}

/*.......................................................................
 * Return the calibration diode settling-time compensation delay for
 * a given integration-period within a fully-described scan.
 *
 * Input:
 *  cnf          const CCBConfig *  The configuration of the CCB.
 *  integration    unsigned long    The sequential number of the
 *                                  integration period within its
 *                                  parent scan, starting at zero for
 *                                  the first integration period.
 * Input/Output:
 *  dt               CCBInterval *  On output, the calibration diode
 *                                  delay will be assigned to *dt.
 * Output:
 *  return                   int    0 - OK.
 *                                  1 - Error (see errno).
 */
int ccb_cal_diode_delay(const CCBConfig *cnf, unsigned long integration,
			CCBInterval *dt)
{
  unsigned long delay_a;    /* The delay needed by cal-diode A */
  unsigned long delay_b;    /* The delay needed by cal-diode B */
  unsigned long delay;      /* The combined delay */
  unsigned long offset;     /* The index of the specified integration within */
                            /*  its parent calibration cycle */
  unsigned long now;        /* The index of the calibration step of the */
                            /*  specified integration. */
  unsigned long before;     /* The index of the calibration step of the */
                            /*  integration that preceded the specified */
                            /*  integration. */
  unsigned long sum;        /* A sum of integrations */
  unsigned long cal_period; /* The number of integrations per cal-diode */
                            /*  cycle. */
  CCBPhaseSwitchCnf phase;  /* The CCB phase-switch configuration */
  CCBCalDiodeCnf cal;       /* The CCB calibration diode configuration */
  CCBTimingCnf timing;      /* The CCB timing configuration */
  int i;
/*
 * Check the arguments.
 */
  if(!cnf || !dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_phase_switch_cnf(cnf, &phase) ||
     ccb_get_cal_diode_cnf(cnf, &cal) ||
     ccb_get_timing_cnf(cnf, &timing))
    return 1;
/*
 * Total-up the number of integrations in a complete calibration cycle.
 */
  cal_period = 0;
  for(i=0; i<cal.ncal; i++)
    cal_period += cal.diode_times[i];
/*
 * Determine the offset of the specified integration within its
 * parent calibration cycle.
 */
  offset = integration % cal_period;
/*
 * Find out the index of the calibration step that this corresponds to.
 */
  for(sum=0,now=0; sum + cal.diode_times[now] <= offset; now++)
    sum += cal.diode_times[now];
/*
 * Get the index of the calibration step of the preceding integration.
 */
  if(offset > sum)  /* Was the integration in the same calibration step? */
    before = now;
  else if(now > 0)  /* Was the preceding cal-step within the same cycle? */
    before = now - 1;
  else              /* The preceding step was the last of the previous cycle */
    before = cal.ncal - 1;
/*
 * Compute the delay needed by calibration diode A.
 */
  if(!(cal.driven_diodes & CCB_CAL_DIODE_A) || /* If !driven or */
     (integration!=0 &&                        /*  no state change. */
      CCB_CAL_DIODE_A & ~(cal.diode_states[now] ^ cal.diode_states[before]))) 
    delay_a = 0;
  else if(cal.diode_states[now] & CCB_CAL_DIODE_A &&    /* Switch on*/
	  (integration==0 || cal.diode_states[before] ^ CCB_CAL_DIODE_A))
    delay_a = timing.diode_rise_dt;
  else                            /* Switch off */
    delay_a = timing.diode_fall_dt;
/*
 * Compute the delay needed by calibration diode B.
 */
  if(!(cal.driven_diodes & CCB_CAL_DIODE_B) || /* If !driven or */
     (integration!=0 &&                        /*  no state change. */
      CCB_CAL_DIODE_B & ~(cal.diode_states[now] ^ cal.diode_states[before]))) 
    delay_b = 0;
  else if(cal.diode_states[now] & CCB_CAL_DIODE_B &&    /* Switch on*/
	  (integration==0 || cal.diode_states[before] ^ CCB_CAL_DIODE_B))
    delay_b = timing.diode_rise_dt;
  else                            /* Switch off */
    delay_b = timing.diode_fall_dt;
/*
 * Get the maximum of the two diode delays.
 */
  delay = delay_a > delay_b ? delay_a : delay_b;
/*
 * For the first integration of the scan, compute the delay needed
 * by the phase switches, and include this in the maximum.
 */
  if(integration==0) {
    unsigned long psw_delay =
      (phase.driven_switches & CCB_ALL_PHASE_SWITCHES) ? /* Either switch? */
	timing.phase_switch_dt : 0;
    if(psw_delay > delay)
      delay = psw_delay;
  };
/*
 * Convert this to a time interval.
 */
  ccb_clock_interval(delay, dt);
  return 0;
}

/*.......................................................................
 * Return the number of measurements per phase-switch cycle, according
 * to the set of phase-switches that are active.
 *
 * Input:
 *  active_switches  unsigned short   (See ccb_set_phase_switch_cnf())
 * Output:
 *  return                 unsigned    The number of measurements per
 *                                     phase-switch cycle.
 */
unsigned ccb_cycle_length(unsigned short active_switches)
{
/*
 * Are both switches active?
 */
  if((active_switches & CCB_ALL_PHASE_SWITCHES) == CCB_ALL_PHASE_SWITCHES)
    return 4;
/*
 * Either one of the switches?
 */
  else if(active_switches & CCB_ALL_PHASE_SWITCHES)
    return 2;
/*
 * Neither of the switches is active.
 */
  return 1;
}

/*.......................................................................
 * Compute the duration of an integration, excluding any calibration
 * diode delays that precede the integration.
 *
 * Input:
 *  cnf       const CCBConfig *  The configuration of the CCB.
 * Input/Output:
 *  dt            CCBInterval *  The integration duration will be
 *                               recorded in *dt.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error (see errno).
 */
int ccb_nocal_duration(const CCBConfig *cnf, CCBInterval *dt)
{
  unsigned long nsample;     /* The number of A/D samples per integration */
  unsigned long period;      /* The period between A/D samples */
  CCBPhaseSwitchCnf phase;   /* The CCB phase-switch configuration */
  CCBTimingCnf timing;       /* The CCB timing configuration */
/*
 * Check the arguments.
 */
  if(!cnf || !dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_phase_switch_cnf(cnf, &phase) ||
     ccb_get_timing_cnf(cnf, &timing))
    return 1;
/*
 * Compute the number of samples per integration.
 */
  nsample = timing.integ_period *
    ccb_cycle_length(phase.active_switches) * phase.samp_per_state;
/*
 * What is the period from the start of one sample to the start of
 * the next (expressed as the equivalent number of CCB hardware
 * clock ticks)?
 */
  period = timing.sample_dt + timing.analog_reset_dt;
/*
 * Convert this to a time interval.
 */
  ccb_clock_interval(period, dt);
/*
 * Multiply the sample period by the number of samples per integration.
 * This constitutes the integration duration.
 */
  if(ccb_scale_interval(dt, nsample, dt))
    return 1;
  return 0;
}

/*.......................................................................
 * Return the predicted duration of a fully-described scan of a given
 * number of integration periods.
 *
 * Input:
 *  cnf          const CCBConfig * The configuration of the CCB.
 *  integrations   unsigned long   The number of integrations within
 *                                 the scan.
 * Input/Output:
 *  dt               CCBInterval * The scan duration will be recorded
 *                                 in *dt.
 * Output:
 *  return                   int   0 - OK.
 *                                 1 - Error (see errno).
 */
int ccb_scan_duration(const CCBConfig *cnf, unsigned long integrations,
		      CCBInterval *dt)
{
  unsigned ncomplete;       /* The number of complete calibration cycles */
  unsigned nextra_integ;    /* The number of integrations beyond the final */
                            /*  complete calibration cycle. */
  unsigned nextra_steps;    /* The number of started calibration steps within */
                            /*  the final calibration cycle. */
  unsigned sum;             /* A sum of integrations */
  unsigned long offset;     /* The starting integration offset of a given */
                            /*  calibration-step. */
  CCBInterval integ_dt;     /* The duration of an integration */
  CCBInterval cal_dt;       /* The duration of a calibration diode delay */
  unsigned long cal_period; /* The number of integrations per cal-diode */
                            /*  cycle. */
  CCBCalDiodeCnf cal;       /* The CCB calibration diode configuration */
  int i;
/*
 * Check the arguments.
 */
  if(!cnf || !dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_cal_diode_cnf(cnf, &cal))
    return 1;
/*
 * Total-up the number of integrations in a complete calibration cycle.
 */
  cal_period = 0;
  for(i=0; i<cal.ncal; i++)
    cal_period += cal.diode_times[i];
/*
 * Compute the duration of an integration, minus calibration diode
 * settling-time delays.
 */
  if(ccb_nocal_duration(cnf, &integ_dt))
    return 1;
/*
 * Scale the duration of an integration by the number of integrations
 * in the scan, and record the result as part of the final scan duration.
 */
  if(ccb_scale_interval(&integ_dt, integrations, dt))
    return 1;
/*
 * How many complete calibration cycles are encompassed by 'integrations'
 * integrations?
 */
  ncomplete = integrations / cal_period;
/*
 * Compute the number of integrations in the final incomplete cycle.
 */
  nextra_integ = integrations - (ncomplete * cal_period);
/*
 * Determine the number of calibration steps started in the final
 * calibration cycle.
 */
  for(sum=0,nextra_steps=0; nextra_integ > sum; nextra_steps++)
    sum += cal.diode_times[nextra_steps];
/*
 * Compute the abnormal cal-diode delay of the first integration.
 */
  if(ccb_cal_diode_delay(cnf, 0, &cal_dt))
    return 1;
/*
 * Add this to the total scan duration.
 */
  if(ccb_add_intervals(dt, &cal_dt, dt))
    return 1;
/*
 * For each step of the repeated calibration cycle, compute the total
 * amount of time taken by the calibration diode settling times of the
 * diodes at the start of the step, and add this to the total time.
 */
  for(offset=0,i=0; i<cal.ncal; offset += cal.diode_times[i], i++) {
/*
 * How many integrations are accumulated during the i'th step of the
 * calibration cycle? Exclude the first integration of the first step,
 * since we already added its contribution to the answer.
 */
    unsigned long n = ncomplete - (i==0) + (i<nextra_steps);
/*
 * Compute the settling time of the diodes for a single integration to
 * reach the cal-diode state of the current part of the cal cycle.
 * Note that setting the 'integration' argument is set to cal.ncal+offset,
 * rather than just 'offset', to avoid the special behavior of i==0, which we
 * already took care of above. Otherwise (cal_period+offset)
 * and (offset) are equivalent, since the cycle repeats every
 * cal_period integrations.
 */
    if(ccb_cal_diode_delay(cnf, cal_period+offset, &cal_dt))
      return 1;
/*
 * Scale this delay by the number of integrations taken in this state.
 */
    if(ccb_scale_interval(&cal_dt, n, &cal_dt))
      return 1;
/*
 * Add the result to the total scan duration.
 */
    if(ccb_add_intervals(dt, &cal_dt, dt))
      return 1;
  };
/*
 * The total scan duration is now recorded in *dt.
 */
  return 0;
}

/*.......................................................................
 * Compute the integration-time of an integration. This is the actual
 * amount of time integrating, and thus excludes cal-diode delays,
 * phase-switching delays and the delay while the analog integrator is
 * being reset.
 *
 * Input:
 *  cnf       const CCBConfig *  The configuration of the CCB.
 * Input/Output:
 *  dt            CCBInterval *  The integration time will be
 *                               recorded in *dt.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error (see errno).
 */
int ccb_integration_time(const CCBConfig *cnf, CCBInterval *dt)
{
  CCBPhaseSwitchCnf phase;   /* The CCB phase-switch configuration */
  CCBTimingCnf timing;       /* The CCB timing configuration */
/*
 * Check the arguments.
 */
  if(!cnf || !dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_phase_switch_cnf(cnf, &phase) ||
     ccb_get_timing_cnf(cnf, &timing))
    return 1;
/*
 * Convert the A/D sampling time from a number of clock ticks, to a
 * physical time.
 */
  ccb_clock_interval(timing.sample_dt, dt);
/*
 * Multiply the sample time by the number of samples per
 * phase-switch state.
 */
  if(ccb_scale_interval(dt, phase.samp_per_state, dt))
    return 1;
/*
 * If either of the phase switches are active, subtract the
 * phase-switch transition delay.
 */
  if(phase.active_switches & CCB_ALL_PHASE_SWITCHES) {
    CCBInterval phase_switch_time;
    ccb_clock_interval(timing.phase_switch_dt, &phase_switch_time);
    if(ccb_subtract_interval(dt, &phase_switch_time, dt))
      return 1;
  };
/*
 * Scale the integration-time of one phase-switch state by the
 * number of states per integration period.
 */
  if(ccb_scale_interval(dt,
	 timing.integ_period * ccb_cycle_length(phase.active_switches), dt))
    return 1;
  return 0;
}

/*.......................................................................
 * Compute the duration of an integration period.
 *
 * Input:
 *  cnf         const CCBConfig *  The configuration of the CCB.
 *  integration   unsigned long    The sequential number of the
 *                                 integration period within its
 *                                 parent scan, starting at zero for
 *                                 the first integration period.
 * Input/Output:
 *  dt              CCBInterval *  The integration duration will be
 *                                 recorded in *dt.
 * Output:
 *  return                  int    0 - OK.
 *                                 1 - Error (see errno).
 */
int ccb_integration_duration(const CCBConfig *cnf, unsigned long integration,
			     CCBInterval *dt)
{
  CCBInterval diode_dt;  /* The calibration settling time delay */
  CCBInterval integ_dt;  /* The duration of the integration */
/*
 * Check those arguments that aren't checked by other functions before they
 * are used.
 */
  if(!cnf || !dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Compute the contribution of the calibration diode transition delay.
 */
  if(ccb_cal_diode_delay(cnf, integration, &diode_dt))
    return 1;
/*
 * Compute the part of the integration duration that doesn't depend on
 * calibration diode switching.
 */
  if(ccb_nocal_duration(cnf, &integ_dt))
    return 1;
/*
 * The total integration duration is the sum of the cal-diode switching
 * delay and the cal-diode-invariant part of the integration duration.
 */
  if(ccb_add_intervals(&diode_dt, &integ_dt, dt))
    return 1;
  return 0;
}
