#ifndef ccbserverlink_h
#define ccbserverlink_h

/*-----------------------------------------------------------------------*
 * This is the public header file of the client portion of the CCB       *
 * communications library.                                               *
 *-----------------------------------------------------------------------*/

/*
 * Get datatypes that are common to both the server and client sides of
 * the CCB communications library.
 */
#include "ccbcommon.h"

#include <sys/types.h>   /* fd_set */
#include <stdarg.h>

/*
 * The CCBServerLink datatype holds the dynamically-allocated resources
 * that CCB server communications with a CCB manager. The contents of
 * the structure are private to the library, so this public header
 * file only includes an opaque declaration.
 */
typedef struct CCBServerLink CCBServerLink;

/*
 * The CCBDriverCmd type encapsulates commands that are sent to the
 * CCB device driver. It's contents are defined later in this file.
 */
typedef struct CCBDriverCmd CCBDriverCmd;

/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to load the CCB device driver.  Note that this function
 * is only called after a preceding call to ccb_unload_driver().
 *
 * Input:
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_DRIVER_LOAD_FN(fn) int (fn)(void *data)

typedef CCB_DRIVER_LOAD_FN(CCBDriverLoadFn);

/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to unload the CCB device driver.  Beware that this
 * function will sometimes be called when no device driver is currently
 * loaded. This should not be treated as an error.
 *
 * Input:
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_DRIVER_UNLOAD_FN(fn) int (fn)(void *data)

typedef CCB_DRIVER_UNLOAD_FN(CCBDriverUnloadFn);


/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to send a command to the CCB device driver.
 *
 * Input:
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 *  cmd  const CCBDriverCmd *  The command to be sent to the driver.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_DRIVER_TELL_FN(fn) int (fn)(void *data, const CCBDriverCmd *cmd)

typedef CCB_DRIVER_TELL_FN(CCBDriverTellFn);

/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to reboot the CCB real-time computer. The function need
 * not wait for the reboot process to actually start.
 *
 * Input:
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_REBOOT_RTC_FN(fn) int (fn)(void *data)

typedef CCB_REBOOT_RTC_FN(CCBRebootRTCFn);

/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to shutdown the CCB real-time computer. The function need
 * not wait for the shutdown process to actually start.
 *
 * Input:
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_SHUTDOWN_RTC_FN(fn) int (fn)(void *data)

typedef CCB_SHUTDOWN_RTC_FN(CCBShutdownRTCFn);

/*
 * Create the resources needed to communicate with a remote CCB manager
 * process.
 */
CCBServerLink *new_CCBServerLink(void *data,
				 CCBDriverLoadFn *load_driver,
				 CCBDriverUnloadFn *unload_driver,
				 CCBDriverTellFn *tell_driver,
				 CCBRebootRTCFn *reboot_rtc,
				 CCBShutdownRTCFn *shutdown_rtc);

/*
 * Shutdown the communications of any connected manager, then return
 * the resources of the specified CCBServerLink object to the system.
 */
CCBServerLink *del_CCBServerLink(CCBServerLink *sl);

/*
 * Augment select() file-descriptor sets according to the socket
 * I/O that the server library is watching for.
 */
int ccb_server_select_args(CCBServerLink *sl, fd_set *rfds, fd_set *wfds,
			   int *maxfd);

/*
 * Perform socket-level I/O on the CCB server sockets.
 */
int ccb_server_communicate(CCBServerLink *ccb, const fd_set *rfds,
			   const fd_set *wfds);

/*
 * Queue an integ-data message to be sent to the current manager.
 */
int ccb_queue_integ_msg(CCBServerLink *sl, const CCBTimeStamp *t,
			unsigned long scan, unsigned long number,
			const unsigned long *values, unsigned nvalues);
/*
 * Queue a monitor-data message to be sent to the current manager.
 */
int ccb_queue_monitor_msg(CCBServerLink *sl, const CCBTimeStamp *t,
			  unsigned long scan, unsigned long number,
			  const unsigned long *values, unsigned nvalues);

/* The following function is used for reporting all CCB server log messages */

int ccb_log_server_msg(CCBServerLink *sl, CCBLogLevel level, unsigned long id,
		       const char *fmt, ...)
#ifdef __GNUC__                          /* Arrange for gcc to check the */
__attribute__ ((format (printf, 4, 5))); /* printf-style format and arguments */
#else
 ;
#endif

/*
 * Record the existence of software and/or hardware faults.
 */
int ccb_record_server_faults(CCBServerLink *sl, unsigned faults);

/*
 * Enumerate all device-driver commands sent by the server.
 */
typedef enum {
  CCB_DRV_INTR_MASK,   /* Enable/disable interrupts */
  CCB_DRV_LINE_MASK,   /* Specify a control-line drive mask to be bitwise */
                       /*  ANDed with the configured set of driven lines. */
  CCB_DRV_ABORT_SCAN,  /* Abort the current scan, and don't start a new one */
  CCB_DRV_CONF_SCAN,   /* Install a new configuration */
  CCB_DRV_STAGE_SCAN,  /* Start a new scan at the specified time */
  CCB_DRV_INTRA_SCAN   /* Start an intra-scan ASAP */
} CCBDriverCmdID;

/*
 * Enumerate the CCB interrupt sources, using one bit per interrupt.
 */
typedef enum {
  CCB_1PPS_INTR=1,     /* The 1-pulse-per-second interrupt */
  CCB_INTEG_INTR=2     /* The integration-done interrupt */
} CCBDrvInterrupt;

/*
 * This is the body of a CCB_DRV_INTR_MASK command.
 */
typedef struct {
  unsigned enable;     /* A bitmask union of CCBDrvInterrupt enumerators */
                       /*  specifying which interrupts to enable. */
} CCBDrvIntrMask;

/*
 * Enumerate the receiver control-lines, using one bit per control-line.
 */
typedef enum {
  CCB_DIODE_A_LINE = 1,   /* Calibration-diode A's control lines */
  CCB_DIODE_B_LINE = 2,   /* Calibration-diode B's control lines */
  CCB_PHASE_A_LINE = 4,   /* Phase-switch A's control line */
  CCB_PHASE_B_LINE = 8,   /* Phase-switch B's control line */
/*
 * All of the above.
 */
  CCB_ALL_LINES = CCB_DIODE_A_LINE | CCB_DIODE_B_LINE | CCB_PHASE_A_LINE |
    CCB_PHASE_B_LINE,
/*
 * None of the above.
 */
  CCB_NO_LINES = 0
} CCBControlLine;

/*
 * This is the body of a CCB_DRV_LINE_MASK command.
 */
typedef struct {
  unsigned lines;  /* A bitmask union of CCBControlLine enumerators */
                   /*  specifying which receiver control lines can */
                   /*  be driven, if so configured. */
} CCBDrvLineMask;

/*
 * This is the body of a CCB_DRV_CONF_SCAN command.
 */
typedef struct {
  CCBPhaseSwitchCnf phase;  /* The phase-switch configuration */
  CCBCalDiodeCnf cal;       /* The calibration diode configuration */
  CCBTimingCnf timing;
} CCBDrvConfScan;

/*
 * This is the body of the CCB_DRV_STAGE_SCAN command.
 */
typedef struct {
  unsigned long scan; /* The numeric scan identifier */
  unsigned long mjd;  /* The date on which to start, expressed as a Modified */
                      /*  Julian day number. */
  unsigned long secs; /* The time of day at which to start the scan, */
                      /*  expressed in seconds since 0H UTC on the */
                      /*  day specified in 'mjd'. */
} CCBDrvStageScan;

/*
 * This is the body of the CCB_DRV_INTRA_SCAN command.
 */
typedef struct {
  unsigned long scan; /* The numeric scan identifier */
} CCBDrvIntraScan;

/*
 * Define the container which is used to send commands to the
 * device driver.
 */
struct CCBDriverCmd {
  CCBDriverCmdID type;      /* The type of command being sent */
  union {
    CCBDrvIntrMask intr;    /* type==CCB_DRV_INTR_MASK */
    CCBDrvLineMask line;    /* type==CCB_DRV_LINE_MASK */
    CCBDrvConfScan conf;    /* type==CCB_DRV_CONF_SCAN */
    CCBDrvStageScan stage;  /* type==CCB_DRV_STAGE_SCAN */
    CCBDrvIntraScan intra;  /* type==CCB_DRV_INTRA_SCAN */
  } pars;
};

#endif
