#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "iptable.h"
#include "freelist.h"
#include "ccbprint.h"
#include "input.h"

/*
 * Enumerate the number of fields in an IPV4 address.
 */
enum {NUM_IP_FIELD=4};

/*
 * Each IP address specification is recorded in a list node of
 * the following type.
 */
typedef struct CCBIPSpec CCBIPSpec;
struct CCBIPSpec {
  CCBIPSpec *next;           /* The next address specification in the list */
  short fields[NUM_IP_FIELD];/* The numeric values of each of the fields of */
                             /*  the IP address, with -1 being used to */
                             /*  denote that the field is wildcarded. */
};

/*
 * Define an object that maintains a list of IP addresses.
 */
struct CCBIPTable {
  FreeList *spec_mem;   /* A free-list of CCBIPSpec structures */
  InputStream *stream;  /* A stream for parsing IP addresses */
  CCBIPSpec *head;      /* The list of IP address specifications */
};

static int ip_read_addresses(CCBIPTable *ip, const char *dir,
			     const char *filename, char *errmsg,
			     size_t errdim);
static int cant_read_ip_addresses(CCBIPTable *ip);
static int ip_read_entry(CCBIPTable *ip, InputStream *stream, char *errmsg,
			 size_t errdim, int nfield, short *fields);

/*
 * Specify how many new specification nodes should be allocated whenever
 * the freelist becomes exhausted.
 */
#define CCB_IP_SPEC_BLOCKING 100

/*.......................................................................
 * Allocate a new IP-address table, and fill it from the contents of a
 * specified file.
 *
 * Input:
 *  dir       const char *  The directory in which to search for the file.
 *  filename  const char *  The name of a file that contains a list of
 *                          numeric IP addresses, with one address per
 *                          line. Fields within each IP address can be
 *                          wildcarded by writing an asterix in place
 *                          of the number that would otherwise go there.
 *  errmsg          char *  A buffer in which to place an error message
 *                          if this function returns NULL.
 *  errdim        size_t    The allocated size of errmsg[].
 * Output:
 *  return    CCBIPTable *  The new object, or NULL on error.
 */
CCBIPTable *new_CCBIPTable(const char *dir, const char *filename,
			   char *errmsg, size_t errdim)
{
  CCBIPTable *ip;  /* The object to be returned */
/*
 * Check the arguments.
 */
  if(!filename) {
    ccb_snprintf(errmsg, errdim, "new_CCBIPTable: NULL argument(s).\n");
    errno = EINVAL;
    return NULL;
  };
/*
 * Allocate the container.
 */
  ip = malloc(sizeof(CCBIPTable));
  if(!ip) {
    ccb_snprintf(errmsg, errdim, "new_CCBIPTable: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBIPTable().
 */
  ip->spec_mem = NULL;
  ip->stream = NULL;
  ip->head = NULL;
/*
 * Allocate the free-list of CCBIPSpec nodes.
 */
  ip->spec_mem = _new_FreeList(sizeof(CCBIPSpec), CCB_IP_SPEC_BLOCKING);
  if(!ip->spec_mem) {
    ccb_snprintf(errmsg, errdim, "new_CCBIPTable: Insufficient memory.\n");
    return del_CCBIPTable(ip);
  };
/*
 * Allocate an input stream to use to parse IP addresses.
 */
  ip->stream = new_InputStream();
  if(!ip->stream) {
    ccb_snprintf(errmsg, errdim, "new_CCBIPTable: Insufficient memory.\n");
    return del_CCBIPTable(ip);
  };
/*
 * Read the file of IP address specifications.
 */
  if(ip_read_addresses(ip, dir, filename, errmsg, errdim))
    return del_CCBIPTable(ip);
  return ip;
}

/*.......................................................................
 * Delete a CCBIPTable object.
 *
 * Input:
 *  ip   CCBIPTable *  The object to be deleted.
 * Output:
 *  return CCBIPTable *  The deleted object (always NULL).
 */
CCBIPTable *del_CCBIPTable(CCBIPTable *ip)
{
  if(ip) {
    ip->spec_mem = _del_FreeList(ip->spec_mem, 1);
    ip->stream = del_InputStream(ip->stream);
    free(ip);
  };
  return NULL;
}

/*.......................................................................
 * Read a list of IP address specifications from a specified file.
 *
 * Input:
 *  ip        CCBIPTable *   The object in which to record the list.
 *  dir       const char *  The directory in which to search for the file.
 *  filename  const char *  The name of a file that contains a list of
 *                          numeric IP addresses, with one address per
 *                          line. Fields within each IP address can be
 *                          wildcarded by writing an asterix in place
 *                          of the number that would otherwise go there.
 *  errmsg          char *  A buffer in which to place an error message
 *                          if this function returns NULL.
 *  errdim        size_t    The allocated size of errmsg[].
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int ip_read_addresses(CCBIPTable *ip, const char *dir,
			     const char *filename, char *errmsg, size_t errdim)
{
  short fields[NUM_IP_FIELD];   /* The decoded address specification */
  CCBIPSpec *spec;              /* A the new IP address specification */
  int i;
/*
 * Attempt to open the specified file, and attach the above input
 * stream to it.
 */
  if(open_FileInputStream(ip->stream, dir, filename)) {
    last_input_error(ip->stream, errmsg, errdim);
    return cant_read_ip_addresses(ip);
  };
/*
 * Read and decode one entry at a time, until the end of the file
 * is reached, or an error occurs.
 */
  do {
/*
 * Read the next specification.
 */
    if(ip_read_entry(ip, ip->stream, errmsg, errdim, NUM_IP_FIELD, fields))
      return cant_read_ip_addresses(ip);
/*
 * Allocate a new node for the list of address-specifications.
 */
    spec = _new_FreeListNode(ip->spec_mem);
    if(!spec) {
      ccb_snprintf(errmsg, errdim,
		   "Insufficient memory to record contents of '%s'", filename);
      return cant_read_ip_addresses(ip);
    };
/*
 * Initialize the specification node.
 */
    spec->next = NULL;
    for(i=0; i<NUM_IP_FIELD; i++)
      spec->fields[i] = fields[i];
/*
 * Prepend the new specification to the list.
 */
    spec->next = ip->head;
    ip->head = spec;
/*
 * Go back to read the next specification unless the end of the file
 * has been reached.
 */
  } while(input_peek_char(ip->stream) != EOF);
/*
 * Clean up.
 */
  close_InputStream(ip->stream);
  return 0;
}

/*.......................................................................
 * This is the private error-return function of ip_read_addresses(). It
 * closes the file, then returns the error-return code of 
 * ip_read_addresses().
 */
static int cant_read_ip_addresses(CCBIPTable *ip) 
{
  close_InputStream(ip->stream);
  return 1;
}

/*.......................................................................
 * Read a single white-space delimited IP address specification from a
 * given stdio stream, and record it in a table.
 *
 * Input:
 *  ip        CCBIPTable *  The object in which to record the list.
 *  stream   InputStream *  The input stream to read from.
 *  errmsg          char *  A buffer in which to place an error message
 *                          if this function returns 1.
 *  errdim        size_t    The allocated size of errmsg[].
 *  nfield           int    The number of fields to read.
 * Input/Output
 *  fields         short *  The 'nfield' field values will be written
 *                          in the array pointed to by this argument.
 *                          Wildcarded field entries are represented
 *                          as -1.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int ip_read_entry(CCBIPTable *ip, InputStream *stream, char *errmsg,
			 size_t errdim, int nfield, short *fields)
{
  int c;                          /* A character from the input stream */
  int i;
/*
 * Skip leading white-space and comments.
 */
  if(input_skip_white(stream, 0)) {
    last_input_error(stream, errmsg, errdim);
    return 1;
  };
/*
 * Find out what character terminated the section of white-space
 * and/or comments.
 */
  c = input_peek_char(stream);
/*
 * Did we hit the end of the file?
 */
  if(c==EOF)
    return 0;
/*
 * Decode the individual fields of the specification.
 */
  for(i=0; i<nfield; i++) {
/*
 * Get the first character of the field.
 */
    c = input_peek_char(stream);
/*
 * Is the character a wildcard?
 */
    if(c == '*') {
      input_skip_char(stream);
      fields[i] = -1;
/*
 * Is the field a normal integer field?
 */
    } else if(isdigit((int)(unsigned char) c)) {
/*
 * Read the integer as an unsigned long.
 */
      unsigned long ul;
      if(input_ulong(stream, 0, &ul)) {
	last_input_error(stream, errmsg, errdim);
	return 1;
      };
/*
 * The components of IPV4 addresses are unsigned 8-bit integers.
 */
      if(ul > 255) {
	input_error(stream, "Invalid IP address specification");
	last_input_error(stream, errmsg, errdim);
	return 1;
      };
/*
 * Record the new component.
 */
      fields[i] = (short) ul;
/*
 * Anything else than an integer or a wildcard character is illegal.
 */
    } else {
      input_error(stream, "Unexpected characters in specification");
      last_input_error(stream, errmsg, errdim);
      return 1;
    };
/*
 * All but the last component of the address must be followed by a
 * period.
 */
    if(i < nfield-1) {
/*
 * Check for the period.
 */
      if(input_peek_char(stream) != '.') {
	input_error(stream, "Missing period after IP address component");
	last_input_error(stream, errmsg, errdim);
	return 1;
      };
/*
 * Skip the period.
 */
      input_skip_char(stream);
    };
  };
/*
 * Trailing characters constitute an error.
 */
  c = input_peek_char(stream);
  if(c != EOF && !isspace((int)(unsigned char) c)) {
    input_error(stream, "Unexpected characters at end of valid specification");
    last_input_error(stream, errmsg, errdim);
    return 1;
  };
  return 0;
}

/*.......................................................................
 * Check wether a given IP address is covered by a specified table
 * of IP addresses, and return true (non-zero) if so.
 *
 * Input:
 *  ip    CCBIPTable *  The table of IP addresses.
 *  addr        char *  The numeric IP address to validate.
 * Output:
 *  return       int    0 - Reject this address.
 *                      1 - Accept this address.
 */
int ccb_valid_manager_address(CCBIPTable *ip, const char *addr)
{
  CCBIPSpec *spec;            /* A node in the list of address specifications */
  short fields[NUM_IP_FIELD]; /* The decoded address specification */
  int i;
/*
 * Attach an input stream to the buffer.
 */
  if(open_StringInputStream(ip->stream, 0, addr))
    return 0;
/*
 * Decompose the textually specified address into its numeric components.
 */
  if(ip_read_entry(ip, ip->stream, NULL, 0, NUM_IP_FIELD, fields)) {
    close_InputStream(ip->stream);
    return 0;
  };
/*
 * We have finished parsing the address.
 */
  close_InputStream(ip->stream);
/*
 * Compare the address to each of the specifications.
 */
  for(spec=ip->head; spec; spec=spec->next) {
/*
 * Search for any non-matching field in the latest specification.
 */
    for(i=0; i<NUM_IP_FIELD; i++) {
      if(spec->fields[i] >= 0 && fields[i] != spec->fields[i])
	break;
    };
/*
 * If all fields matched, the address is valid.
 */
    if(i==NUM_IP_FIELD)
      return 1;
  };
/*
 * None of the addresses matched.
 */
  return 0;
}
